import numpy as np
import math
import shap

from .base_plotting import PlotStructure
from ..common.contrib_utils import combine_like_features
import matplotlib.pyplot as plt
from .dependence import dependence_plot
from matplotlib.lines import Line2D
import matplotlib
import seaborn as sns

import re
from shap.plots import colors


def format_value(s, format_str):
    """Strips trailing zeros and uses a unicode minus sign."""
    if not issubclass(type(s), str):
        s = format_str % s
    s = re.sub(r"\.?0+$", "", s)
    if s[0] == "-":
        s = "\u2212" + s[1:]
    return s


def waterfall(
    data,
    key,
    model_name,
    features,
    ax=None,
    fig=None,
    display_feature_names={},
    display_units={},
    label_fontsize=8,
    **kwargs,
):

    """Plots an explantion of a single prediction as a waterfall plot.
    The SHAP value of a feature represents the impact of the evidence provided by that feature on the model's
    output. The waterfall plot is designed to visually display how the SHAP values (evidence) of each feature
    move the model output from our prior expectation under the background data distribution, to the final model
    prediction given the evidence of all the features. Features are sorted by the magnitude of their SHAP values
    with the smallest magnitude features grouped together at the bottom of the plot when the number of features
    in the models exceeds the max_display parameter.

    Parameters
    ----------
    shap_values : Explanation
        A one-dimensional Explanation object that contains the feature values and SHAP values to plot.
    max_display : str
        The maximum number of features to plot.
    show : bool
        Whether matplotlib.pyplot.show() is called before returning. Setting this to False allows the plot
        to be customized further after it has been created.
    """
    max_display = kwargs.get("max_display")
    all_features = data.attrs["feature_names"]
    feature_names = np.array([display_feature_names.get(f, f) for f in features])
    units = np.array([display_units.get(f, "") for f in features])

    vars_c = [f"{var}_contrib" for var in features if "Bias" not in var]
    vars_val = [f"{var}_val" for var in features if "Bias" not in var]

    values = data.loc[key].loc[model_name, vars_c]
    features = data.loc[key].loc[model_name, vars_val]
    base_values = data.loc[key].loc[model_name, "Bias_contrib"]

    lower_bounds = getattr(values, "lower_bounds", None)
    upper_bounds = getattr(values, "upper_bounds", None)

    # init variables we use for tracking the plot locations
    num_features = min(max_display, len(values))
    row_height = 0.5
    rng = range(num_features - 1, -1, -1)
    order = np.argsort(-np.abs(values))
    pos_lefts = []
    pos_inds = []
    pos_widths = []
    pos_low = []
    pos_high = []
    neg_lefts = []
    neg_inds = []
    neg_widths = []
    neg_low = []
    neg_high = []
    loc = base_values + values.sum()
    yticklabels = ["" for i in range(num_features + 1)]

    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features == len(values):
        num_individual = num_features
    else:
        num_individual = num_features - 1

    # compute the locations of the individual features and plot the dashed connecting lines
    for i in range(num_individual):
        sval = values[order[i]]
        loc -= sval
        if sval >= 0:
            pos_inds.append(rng[i])
            pos_widths.append(sval)
            if lower_bounds is not None:
                pos_low.append(lower_bounds[order[i]])
                pos_high.append(upper_bounds[order[i]])
            pos_lefts.append(loc)
        else:
            neg_inds.append(rng[i])
            neg_widths.append(sval)
            if lower_bounds is not None:
                neg_low.append(lower_bounds[order[i]])
                neg_high.append(upper_bounds[order[i]])
            neg_lefts.append(loc)
        if num_individual != num_features or i + 4 < num_individual:
            ax.plot(
                [loc, loc],
                [rng[i] - 1 - 0.4, rng[i] + 0.4],
                color="#bbbbbb",
                linestyle="--",
                linewidth=0.5,
                zorder=-1,
            )

        if features is None:
            yticklabels[rng[i]] = feature_names[order[i]]
        else:
            if abs(features[order[i]]) < 1:
                fmt = "%0.03f"
            elif abs(features[order[i]]) > 10:
                fmt = "%0.f"
            else:
                fmt = "%0.02f"
            yticklabels[rng[i]] = (
                format_value(features[order[i]], fmt)
                + " "
                + units[order[i]]
                + " = "
                + feature_names[order[i]]
            )

    # add a last grouped feature to represent the impact of all the features we didn't show
    if num_features < len(values):
        yticklabels[0] = "%d other features" % (len(values) - num_features + 1)
        remaining_impact = base_values - loc
        if remaining_impact < 0:
            pos_inds.append(0)
            pos_widths.append(-remaining_impact)
            pos_lefts.append(loc + remaining_impact)
            c = colors.red_rgb
        else:
            neg_inds.append(0)
            neg_widths.append(-remaining_impact)
            neg_lefts.append(loc + remaining_impact)
            c = colors.blue_rgb

    points = (
        pos_lefts
        + list(np.array(pos_lefts) + np.array(pos_widths))
        + neg_lefts
        + list(np.array(neg_lefts) + np.array(neg_widths))
    )
    dataw = np.max(points) - np.min(points)

    # draw invisible bars just for sizing the axes
    label_padding = np.array([0.1 * dataw if w < 1 else 0 for w in pos_widths])
    ax.barh(
        pos_inds,
        np.array(pos_widths) + label_padding + 0.02 * dataw,
        left=np.array(pos_lefts) - 0.01 * dataw,
        color=colors.red_rgb,
        alpha=0,
    )
    label_padding = np.array([-0.1 * dataw if -w < 1 else 0 for w in neg_widths])
    ax.barh(
        neg_inds,
        np.array(neg_widths) + label_padding - 0.02 * dataw,
        left=np.array(neg_lefts) + 0.01 * dataw,
        color=colors.blue_rgb,
        alpha=0,
    )

    # define variable we need for plotting the arrows
    head_length = 0.08
    bar_width = 0.8
    xlen = ax.get_xlim()[1] - ax.get_xlim()[0]
    xticks = ax.get_xticks()
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    bbox_to_xscale = xlen / width
    hl_scaled = bbox_to_xscale * head_length
    renderer = fig.canvas.get_renderer()

    # draw the positive arrows
    for i in range(len(pos_inds)):
        dist = pos_widths[i]
        arrow_obj = ax.arrow(
            pos_lefts[i],
            pos_inds[i],
            max(dist - hl_scaled, 0.000001),
            0,
            head_length=min(dist, hl_scaled),
            color=colors.red_rgb,
            width=bar_width,
            head_width=bar_width,
        )

        if pos_low is not None and i < len(pos_low):
            ax.errorbar(
                pos_lefts[i] + pos_widths[i],
                pos_inds[i],
                xerr=np.array(
                    [[pos_widths[i] - pos_low[i]], [pos_high[i] - pos_widths[i]]]
                ),
                ecolor=colors.light_red_rgb,
            )

        txt_obj = ax.text(
            pos_lefts[i] + 0.5 * dist,
            pos_inds[i],
            format_value(pos_widths[i], "%+0.02f"),
            horizontalalignment="center",
            verticalalignment="center",
            color="white",
            fontsize=12,
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)

        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width:
            txt_obj.remove()

            txt_obj = ax.text(
                pos_lefts[i] + (5 / 72) * bbox_to_xscale + dist,
                pos_inds[i],
                format_value(pos_widths[i], "%+0.02f"),
                horizontalalignment="left",
                verticalalignment="center",
                color=colors.red_rgb,
                fontsize=label_fontsize,
            )

    # draw the negative arrows
    for i in range(len(neg_inds)):
        dist = neg_widths[i]

        arrow_obj = ax.arrow(
            neg_lefts[i],
            neg_inds[i],
            -max(-dist - hl_scaled, 0.000001),
            0,
            head_length=min(-dist, hl_scaled),
            color=colors.blue_rgb,
            width=bar_width,
            head_width=bar_width,
        )

        if neg_low is not None and i < len(neg_low):
            ax.errorbar(
                neg_lefts[i] + neg_widths[i],
                neg_inds[i],
                xerr=np.array(
                    [[neg_widths[i] - neg_low[i]], [neg_high[i] - neg_widths[i]]]
                ),
                ecolor=colors.light_blue_rgb,
            )

        txt_obj = ax.text(
            neg_lefts[i] + 0.5 * dist,
            neg_inds[i],
            format_value(neg_widths[i], "%+0.02f"),
            horizontalalignment="center",
            verticalalignment="center",
            color="white",
            fontsize=label_fontsize,
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)

        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width:
            txt_obj.remove()

            txt_obj = ax.text(
                neg_lefts[i] - (5 / 72) * bbox_to_xscale + dist,
                neg_inds[i],
                format_value(neg_widths[i], "%+0.02f"),
                horizontalalignment="right",
                verticalalignment="center",
                color=colors.blue_rgb,
                fontsize=label_fontsize,
            )

    # draw the y-ticks twice, once in gray and then again with just the feature names in black
    # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax.set_yticks(
        list(range(num_features)) + list(np.arange(num_features) + 1e-8),
    )
    ax.set_yticklabels(
        yticklabels[:-1] + [l.split("=")[-1] for l in yticklabels[:-1]],
        fontsize=label_fontsize,
    )

    # put horizontal lines for each feature row
    for i in range(num_features):
        ax.axhline(i, color="#cccccc", lw=0.5, dashes=(1, 5), zorder=-1)

    # mark the prior expected value and the model prediction
    ax.axvline(
        base_values,
        0,
        1 / num_features,
        color="#bbbbbb",
        linestyle="--",
        linewidth=0.5,
        zorder=-1,
    )
    fx = base_values + values.sum()
    ax.axvline(fx, 0, 1, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)

    # clean up the main axis
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("none")
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_visible(False)
    # ax.tick_params(labelsize=13)
    # pl.xlabel("\nModel output", fontsize=12)

    # draw the E[f(X)] tick mark
    xmin, xmax = ax.get_xlim()
    ax2 = ax.twiny()
    ax2.set_xlim(xmin, xmax)
    ax2.set_xticks(
        [base_values, base_values + 1e-8]
    )  # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax2.set_xticklabels(
        ["\n$E[f(X)]$", "\n$ = " + format_value(base_values, "%0.03f") + "$"],
        fontsize=label_fontsize,
        ha="left",
    )
    ax2.spines["right"].set_visible(False)
    ax2.spines["top"].set_visible(False)
    ax2.spines["left"].set_visible(False)

    # draw the f(x) tick mark
    ax3 = ax2.twiny()
    ax3.set_xlim(xmin, xmax)
    ax3.set_xticks([base_values + values.sum(), base_values + values.sum() + 1e-8])
    # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks

    ax3.set_xticklabels(
        ["$f(x)$", "$ = " + format_value(fx, "%0.03f") + "$"],
        fontsize=label_fontsize,
        ha="left",
    )

    tick_labels = ax3.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(
        tick_labels[0].get_transform()
        + matplotlib.transforms.ScaledTranslation(-10 / 72.0, 0, fig.dpi_scale_trans)
    )
    tick_labels[1].set_transform(
        tick_labels[1].get_transform()
        + matplotlib.transforms.ScaledTranslation(12 / 72.0, 0, fig.dpi_scale_trans)
    )
    tick_labels[1].set_color("#999999")
    ax3.spines["right"].set_visible(False)
    ax3.spines["top"].set_visible(False)
    ax3.spines["left"].set_visible(False)

    # adjust the position of the E[f(X)] = x.xx label
    tick_labels = ax2.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(
        tick_labels[0].get_transform()
        + matplotlib.transforms.ScaledTranslation(-20 / 72.0, 0, fig.dpi_scale_trans)
    )
    tick_labels[1].set_transform(
        tick_labels[1].get_transform()
        + matplotlib.transforms.ScaledTranslation(
            22 / 72.0, -1 / 72.0, fig.dpi_scale_trans
        )
    )

    tick_labels[1].set_color("#999999")

    # color the y tick labels that have the feature values as gray
    # (these fall behind the black ones with just the feature name)
    tick_labels = ax.yaxis.get_majorticklabels()
    for i in range(num_features):
        tick_labels[i].set_color("#999999")

    return fx, base_values


class PlotFeatureContributions(PlotStructure):
    """
    PlotFeatureContributions handles plotting contribution-based plotting for
    single examples or some subset. This class also handles plotting SHAP-style
    plots, which include summary and dependence plots.
    """

    def __init__(self, BASE_FONT_SIZE=12):
        super().__init__(BASE_FONT_SIZE=BASE_FONT_SIZE, set_seaborn=False)
        sns.set_style("white")
        
    def plot_contributions(
        self,
        data,
        estimator_names,
        features,
        to_only_varname=None,
        display_feature_names={},
        display_units={},
        **kwargs,
    ):
        """
        Plot the results of feature contributions

        Args:
        ---------------
            result : pandas.Dataframe
                a single row/example from the
                result dataframe from tree_interpreter_simple
        """
        kwargs["max_display"] = kwargs.get("max_display", 10)
        estimator_output = kwargs.get("estimator_output", None)

        only_one_model = True if len(estimator_names) == 1 else False
        outer_indexs = list(set([f[0] for f in data.index.values]))

        if estimator_output == "probability" and "non_performance" not in outer_indexs:
            outer_indexs = [
                "Best Hits",
                "Worst False Alarms",
                "Worst Misses",
                "Best Corr. Negatives",
            ]
        elif estimator_output == "raw" and "non_performance" not in outer_indexs:
            outer_indexs = ["Least Error Predictions", "Most Error Predictions"]
        perf_keys = kwargs.get("perf_keys", outer_indexs)
        kwargs["wspace"] = kwargs.get("wspace", 0.75)
        kwargs["hspace"] = kwargs.get("hspace", 0.9)

        n_perf_keys = len(perf_keys)
        n_panels = (
            len(estimator_names)
            if "non_performance" in outer_indexs
            else len(estimator_names) * n_perf_keys
        )

        figsize = (
            (3, 2.5)
            if (n_panels == 1 and "non_performance" not in outer_indexs)
            else (8, 6.5)
        )
        kwargs["figsize"] = kwargs.get("figsize", figsize)
        n_columns = 1
        additional_columns = (
            1
            if (n_panels > 1 and not (estimator_output == "raw" and "non_performance"))
            else 0
        )

        n_columns += additional_columns
        kwargs["n_columns"] = kwargs.get("n_columns", n_columns)

        using_internal_ax = True
        if kwargs.get("ax") is not None:
            using_internal_ax = False
            axes = kwargs.get("ax")
            fig = axes.get_figure()
            n_panels = 1
            kwargs.pop("ax")
        else:
            # create subplots, one for each feature
            fig, axes = self.create_subplots(
                n_panels=n_panels,
                sharex=False,
                sharey=False,
                **kwargs,
            )

        ax_iterator = axes.flat if n_panels > 1 else axes

        # try for all_data/average data
        if "non_performance" in outer_indexs:
            for i, model_name in enumerate(estimator_names):
                if n_panels > 1:
                    ax = ax_iterator[i]
                else:
                    ax = axes
                # self._contribution
                final_pred, bias = waterfall(
                    data=data,
                    features=features,
                    model_name=model_name,
                    to_only_varname=to_only_varname,
                    display_feature_names=display_feature_names,
                    display_units=display_units,
                    key="non_performance",
                    ax=ax,
                    fig=fig,
                    label_fontsize=self.FONT_SIZES["tiny"],
                    **kwargs,
                )

                if not only_one_model:
                    ax.set_title(
                        model_name,
                        alpha=0.8,
                        fontdict={"fontsize": self.FONT_SIZES["teensie"]},
                    )

        else:
            # Hard coded in to maintain correct ordering
            if estimator_output == "probability":
                outer_indexs = [
                    "Best Hits",
                    "Worst False Alarms",
                    "Worst Misses",
                    "Best Corr. Negatives",
                ]
            else:
                outer_indexs = ["Least Error Predictions", "Most Error Predictions"]
            # loop over each model creating one panel per model
            c = 0
            for i, model_name in enumerate(estimator_names):
                perf_keys = kwargs.get("perf_keys", outer_indexs)
                for k, perf_key in enumerate(perf_keys):
                    ax = axes[k, i] if not only_one_model else ax_iterator[c]
                    waterfall(
                        data=data,
                        ax=ax,
                        fig=fig,
                        key=perf_key,
                        model_name=model_name,
                        features=features,
                        display_feature_names=display_feature_names,
                        display_units=display_units,
                        label_fontsize=self.FONT_SIZES["tiny"],
                        **kwargs,
                    )

                    # Add outer indexs as titles if only one model
                    if i == 0 and only_one_model:
                        ax.text(
                            0.5,
                            1.2,
                            perf_key,
                            color="xkcd:darkish blue",
                            ha="center",
                            transform=ax.transAxes,
                            fontsize=self.FONT_SIZES["small"],
                        )

                    # Else add the model names as titles.
                    elif k == 0:
                        ax.set_title(
                            model_name,
                            color="xkcd:darkish blue",
                            fontdict={"fontsize": 20},
                        )

                    c += 1

        if "non_performance" not in outer_indexs and not only_one_model:
            self.set_row_labels(
                labels=outer_indexs,
                axes=axes,
                pos=-1,
                rotation=270,
                pad=1.5,
                fontsize=self.FONT_SIZES["tiny"],
            )

        if "non_performance" in outer_indexs:
            pos = (0.95, 0.05)
        else:
            pos = (1.15, -0.025)

        if using_internal_ax:
            self.add_alphabet_label(
                n_panels, axes, pos=pos, fontsize=self.FONT_SIZES["tiny"]
            )

        return fig, axes

    def plot_shap(
        self,
        shap_values,
        X,
        features,
        plot_type,
        display_feature_names={},
        display_units={},
        feature_values=None,
        target_values=None,
        interaction_index="auto",
        **kwargs,
    ):
        """
        Plot SHAP summary or dependence plot.

        """
        self.display_units = display_units
        self.display_feature_names = display_feature_names

        display_feature_names_list = [
            display_feature_names.get(f, f) for f in self.feature_names
        ]

        if plot_type == "summary":
            shap.summary_plot(
                shap_values,
                features=X,
                feature_names=display_feature_names_list,
                max_display=15,
                plot_type="dot",
                alpha=1,
                show=False,
                sort=True,
            )

        elif plot_type == "dependence":
            # Set up the font sizes for matplotlib
            self.display_feature_names = display_feature_names
            left_yaxis_label = "SHAP values (%)\n(Feature Contributions)"
            n_panels = len(features)
            if n_panels <= 6:
                figsize = (8, 5)
            else:
                figsize = (10, 8)

            using_internal_ax = True
            if kwargs.get("ax") is not None:
                using_internal_ax = False
                axes = kwargs.get("ax")
                fig = axes.get_figure()
                n_panels = 1
                kwargs.pop("ax")
            else:
                fig, axes = self.create_subplots(
                    n_panels=n_panels,
                    sharex=False,
                    sharey=False,
                    figsize=figsize,
                    wspace=0.4,
                    hspace=0.5,
                )

            ax_iterator = self.axes_to_iterator(n_panels, axes)
            nticks = 5 if n_panels < 10 else 3
            
            for ax, feature in zip(ax_iterator, features):
                dependence_plot(
                    feature=feature,
                    shap_values=shap_values,
                    X=X,
                    display_feature_names=display_feature_names_list,
                    interaction_index=interaction_index,
                    target_values=target_values,
                    ax=ax,
                    fig=fig,
                    **kwargs,
                )

                if using_internal_ax:
                    self.set_n_ticks(ax, nticks)
                    if n_panels < 10:
                        self.set_minor_ticks(ax)
                    self.set_axis_label(
                        ax, xaxis_label="".join(feature), yaxis_label=""
                    )
                    ax.axhline(
                        y=0.0, color="k", alpha=0.8, linewidth=0.8, linestyle="dashed"
                    )
                    ###ax.set_yticks(self.calculate_ticks(ax=ax, nticks=5, center=False))
                    self.set_n_ticks(ax, nticks)
                    ax.tick_params(axis="both", labelsize=8)
                    vertices = ax.collections[0].get_offsets()
                    self._to_sci_notation(
                        ax=ax,
                        ydata=vertices[:, 1],
                        xdata=vertices[:, 0],
                        colorbar=False,
                    )

            if using_internal_ax:
                major_ax = self.set_major_axis_labels(
                    fig,
                    xlabel=None,
                    ylabel_left=left_yaxis_label,
                    labelpad=25,
                    **kwargs,
                )

                self.add_alphabet_label(n_panels, axes)

            return fig, axes



# =========================================================
# Libraries for the feature contribution computations.
# =========================================================
import pandas as pd


def get_indices_based_on_performance(estimator, X, y, estimator_output, n_samples=None):
    """
     Determines the best hits, worst false alarms, worst misses, and best
     correct negatives using the data provided during initialization.

     Args:
     ------------------
          estimator : The estimator to process
          n_samples: number of "best/worst" X to return. If None,
              the routine uses the whole dataset

    Return:
          a dictionary containing the indices of each of the 4 categories
          listed above
    """
    if len(y) < 1:
        raise ValueError(
            "y is empty! User likely did not initialize ExplainToolkit with 'y'"
        )

    # default is to use all X
    if n_samples is None:
        n_samples = X.shape[0]

    # make sure user didn't goof the input
    if n_samples <= 0:
        print("n_samples less than or equals 0. Defaulting back to all")
        n_samples = X.shape[0]

    if estimator_output == "probability":
        predictions = estimator.predict_proba(X)[:, 1]
    elif estimator_output == "raw":
        predictions = estimator.predict(X)

    diff = y - predictions
    data = {"y": y, "predictions": predictions, "diff": diff}
    df = pd.DataFrame(data)

    if estimator_output == "probability":
        nonevent_X = df[y == 0]
        event_X = df[y == 1]

        event_X_sorted_indices = event_X.sort_values(
            by="diff", ascending=True
        ).index.values

        nonevent_X_sorted_indices = nonevent_X.sort_values(
            by="diff", ascending=False
        ).index.values

        best_hit_indices = event_X_sorted_indices[:n_samples].astype(int)
        worst_miss_indices = event_X_sorted_indices[-n_samples:][::-1].astype(int)

        best_corr_neg_indices = nonevent_X_sorted_indices[:n_samples].astype(int)
        worst_false_alarm_indices = nonevent_X_sorted_indices[-n_samples:][::-1].astype(
            int
        )

        sorted_dict = {
            "Best Hits": best_hit_indices,
            "Worst Misses": worst_miss_indices,
            "Worst False Alarms": worst_false_alarm_indices,
            "Best Corr. Negatives": best_corr_neg_indices,
        }

    else:
        X_sorted_indices = df.sort_values(by="diff", ascending=True).index.values

        least_error_indices = X_sorted_indices[:n_samples].astype(int)
        most_error_indices = X_sorted_indices[-n_samples:].astype(int)

        sorted_dict = {
            "Least Error Predictions": least_error_indices,
            "Most Error Predictions": most_error_indices,
        }

    return sorted_dict


def avg_and_sort_contributions(contrib_dict, feature_val_dict):
    """
    Get the mean value (of data for a predictory) and contribution from
    each predictor and sort"

    Args:
    -----------
        the_dict: dictionary to process
        performance_dict: if using performance based apporach, this should be
            the dictionary with corresponding indices

    Return:

        a dictionary of mean values and contributions
    """
    avg_contrib_dict = {}
    avg_feature_val_dict = {}

    # for hits, misses, etc.
    for key in contrib_dict.keys():
        contrib_df = contrib_dict[key]
        feature_val_df = feature_val_dict[key]

        contrib_series = contrib_df.mean(axis=0)
        feature_val_series = feature_val_df.mean(axis=0)
        feature_val_series["Bias"] = 0.0

        indices = contrib_series.abs().sort_values(ascending=False).index

        sorted_contrib_df = contrib_series.reindex(indices)
        sorted_feature_val_df = feature_val_series.reindex(indices)

        top_contrib = {
            var: contrib_series[var] for var in list(sorted_contrib_df.index)
        }
        top_values = {
            var: feature_val_series[var] for var in list(sorted_feature_val_df.index)
        }

        avg_contrib_dict[key] = top_contrib
        avg_feature_val_dict[key] = top_values

    return avg_contrib_dict, avg_feature_val_dict


def combine_like_features(contrib, varnames):
    """
    Combine the contributions of like features. E.g.,
    multiple statistics of a single variable
    """
    duplicate_vars = {}
    for var in varnames:
        duplicate_vars[var] = [idx for idx, v in enumerate(varnames) if v == var]

    new_contrib = []
    new_varnames = []
    for var in list(duplicate_vars.keys()):
        idxs = duplicate_vars[var]
        new_varnames.append(var)
        new_contrib.append(np.array(contrib)[idxs].sum())

    return new_contrib, new_varnames



import numpy as np
import xarray as xr
import pandas as pd
from collections import ChainMap
from statsmodels.distributions.empirical_distribution import ECDF
from scipy.stats import t


class MissingFeaturesError(Exception):
    """ Raised when features are missing. 
        E.g., All features are require for 
        IAS or MEC
    """
    def __init__(self, estimator_name, missing_features):
        self.message = f"""ALE for {estimator_name} was not computed for all features. 
                        These features were missing: {missing_features}"""
        super().__init__(self.message)
    

def check_all_features_for_ale(ale, estimator_names, features):
    """ Is there ALE values for each feature """
    data_vars = ale.data_vars
    for estimator_name in estimator_names:
        _list = [True if f'{f}__{estimator_name}__ale' in data_vars else False for f in features]
        if not all(_list):
            missing_features = np.array(features)[np.where(~np.array(_list))[0]]
            raise MissingFeaturesError(estimator_name, missing_features) 


def flatten_nested_list(list_of_lists):
    """Turn a list of list into a single, flatten list"""
    all_elements_are_lists = all([is_list(item) for item in list_of_lists])
    if not all_elements_are_lists:
        new_list_of_lists = []
        for item in list_of_lists:
            if is_list(item):
                new_list_of_lists.append(item)
            else:
                new_list_of_lists.append([item])
        list_of_lists = new_list_of_lists

    return [item for elem in list_of_lists for item in elem]


def is_dataset(data):
    return isinstance(data, xr.Dataset)


def is_dataframe(data):
    return isinstance(data, pd.DataFrame)


def check_is_permuted(X, X_permuted):
    permuted_features = []
    for f in X.columns:
        if not np.array_equal(X.loc[:, f], X_permuted.loc[:, f]):
            permuted_features.append(f)

    return permuted_features


def is_correlated(corr_matrix, feature_pairs, rho_threshold=0.8):
    """
    Returns dict where the key are the feature pairs and the items
    are booleans of whether the pair is linearly correlated above the
    given threshold.
    """
    results = {}
    for pair in feature_pairs:
        f1, f2 = pair.split("__")
        corr = corr_matrix[f1][f2]
        results[pair] = round(corr, 3) >= rho_threshold
    return results


def is_fitted(estimator):
    """
    Checks if a scikit-learn estimator/transformer has already been fit.


    Parameters
    ----------
    estimator: scikit-learn estimator (e.g. RandomForestClassifier)
        or transformer (e.g. MinMaxScaler) object


    Returns
    -------
    Boolean that indicates if ``estimator`` has already been fit (True) or not (False).
    """

    attrs = [v for v in vars(estimator) if v.endswith("_") and not v.startswith("__")]

    return len(attrs) != 0


def determine_feature_dtype(X, features):
    """
    Determine if any features are categorical.
    """
    feature_names = list(X.columns)
    non_cat_features = []
    cat_features = []
    for f in features:
        if f not in feature_names:
            raise KeyError(f"'{f}' is not a valid feature.")

        if str(X.dtypes[f]) == "category":
            cat_features.append(f)
        else:
            non_cat_features.append(f)

    return non_cat_features, cat_features


def cartesian(array, out=None):
    """Generate a cartesian product of input array.
    Parameters

    Codes comes directly from sklearn/utils/extmath.py
    ----------
    array : list of array-like
        1-D array to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.
    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(array)) containing cartesian products
        formed of input array.
    X
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])
    """
    array = [np.asarray(x) for x in array]
    shape = (len(x) for x in array)
    dtype = array[0].dtype

    ix = np.indices(shape)
    ix = ix.reshape(len(array), -1).T

    if out is None:
        out = np.empty_like(ix, dtype=dtype)

    for n, arr in enumerate(array):
        out[:, n] = array[n][ix[:, n]]

    return out


def to_dataframe(results, estimator_names, feature_names):
    """
    Convert the feature contribution results to a pandas.DataFrame
    with nested indexing.
    """
    # results[0] = dict of avg. contributions per estimator
    # results[1] = dict of avg. feature values per estimator
    contrib_names = feature_names.copy()
    contrib_names += ["Bias"]

    nested_key = results[0][estimator_names[0]].keys()

    dframes = []
    for key in nested_key:
        data = []
        for name in estimator_names:
            contribs_dict = results[0][name][key]
            vals_dict = results[1][name][key]
            data.append(
                [contribs_dict[f] for f in contrib_names]
                + [vals_dict[f] for f in feature_names]
            )
        column_names = [f + "_contrib" for f in contrib_names] + [
            f + "_val" for f in feature_names
        ]
        df = pd.DataFrame(data, columns=column_names, index=estimator_names)
        dframes.append(df)

    result = pd.concat(dframes, keys=list(nested_key))

    return result


def to_xarray(data):
    """Converts data dict to xarray.Dataset"""
    ds = xr.Dataset(data)
    return ds


def is_str(a):
    """Check if argument is a string"""
    return isinstance(a, str)


def is_list(a):
    """Check if argument is a list"""
    return isinstance(a, list)


def to_list(a):
    """Convert argument to a list"""
    return [a]


def is_tuple(a):
    """Check if argument is a tuple"""
    return isinstance(a, tuple)


def is_valid_feature(features, official_feature_list):
    """Check if a feature is valid"""
    for f in features:
        if isinstance(f, tuple):
            for sub_f in f:
                if sub_f not in official_feature_list:
                    raise Exception(f"Feature {sub_f} is not a valid feature!")
        else:
            if f not in official_feature_list:
                raise Exception(f"Feature {f} is not a valid feature!")


def is_classifier(estimator):
    """Return True if the given estimator is (probably) a classifier.
    Parameters

    Function from base.py in sklearn
    ----------
    estimator : object
        Estimator object to test.
    Returns
    -------
    out : bool
        True if estimator is a classifier and False otherwise.
    """
    return getattr(estimator, "_estimator_type", None) == "classifier"


def is_regressor(estimator):
    """Return True if the given estimator is (probably) a regressor.
    Parameters

    Functions from base.py in sklearn
    ----------
    estimator : object
        Estimator object to test.
    Returns
    -------
    out : bool
        True if estimator is a regressor and False otherwise.
    """
    return getattr(estimator, "_estimator_type", None) == "regressor"


def is_all_dict(alist):
    """Check if every element of a list are dicts"""
    return all([isinstance(l, dict) for l in alist])


def compute_bootstrap_indices(X, subsample=1.0, n_bootstrap=1, seed=90):
    """
    Routine to generate the indices for bootstrapped X.

    Args:
    ----------------
        X : pandas.DataFrame, numpy.array
        subsample : float or integer
        n_bootstrap : integer

    Return:
    ----------------
        bootstrap_indices : list
            list of indices of the size of subsample or subsample*len(X)
    """
    base_random_state = np.random.RandomState(seed=seed)
    random_num_set = base_random_state.choice(10000, size=n_bootstrap, replace=False)
    random_states = [np.random.RandomState(s) for s in random_num_set]

    n_samples = len(X)
    size = int(n_samples * subsample) if subsample <= 1.0 else subsample

    bootstrap_indices = [
        random_state.choice(range(n_samples), size=size).tolist()
        for random_state in random_states
    ]
    return bootstrap_indices


def merge_dict(dicts):
    """Merge a list of dicts into a single dict"""
    return dict(ChainMap(*dicts))


def merge_nested_dict(dicts):
    """
    Merge a list of nested dicts into a single dict
    """
    merged_dict = {}
    for d in dicts:
        for key in d.keys():
            for subkey in d[key].keys():
                if key not in list(merged_dict.keys()):
                    merged_dict[key] = {subkey: {}}
                merged_dict[key][subkey] = d[key][subkey]

    return merged_dict


def is_outlier(points, thresh=3.5):
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.

    Parameters:
    -----------
        points : An numobservations by numdimensions array of observations
        thresh : The modified z-score to use as a threshold. Observations with
            a modified z-score (based on the median absolute deviation) greater
            than this value will be classified as outliers.

    Returns:
    --------
        mask : A numobservations-length boolean array.

    References:
    ----------
        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.median(points, axis=0)
    diff = np.sum((points - median) ** 2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)

    modified_z_score = 0.6745 * diff / med_abs_deviation

    return modified_z_score > thresh


def cmds(D, k=2):
    """Classical multidimensional scaling

    Theory and code references:
    https://en.wikipedia.org/wiki/Multidimensional_scaling#Classical_multidimensional_scaling
    http://www.nervouscomputer.com/hfs/cmdscale-in-python/
    Arguments:
    D -- A squared matrix-like object (array, DataFrame, ....), usually a distance matrix
    """

    n = D.shape[0]
    if D.shape[0] != D.shape[1]:
        raise Exception("The matrix D should be squared")
    if k > (n - 1):
        raise Exception("k should be an integer <= D.shape[0] - 1")

    # (1) Set up the squared proximity matrix
    D_double = np.square(D)
    # (2) Apply double centering: using the centering matrix
    # centering matrix
    center_mat = np.eye(n) - np.ones((n, n)) / n
    # apply the centering
    B = -(1 / 2) * center_mat.dot(D_double).dot(center_mat)
    # (3) Determine the m largest eigenvalues
    # (where m is the number of dimensions desired for the output)
    # extract the eigenvalues
    eigenvals, eigenvecs = np.linalg.eigh(B)
    # sort descending
    idx = np.argsort(eigenvals)[::-1]
    eigenvals = eigenvals[idx]
    eigenvecs = eigenvecs[:, idx]
    # (4) Now, X=eigenvecs.dot(eigen_sqrt_diag),
    # where eigen_sqrt_diag = diag(sqrt(eigenvals))
    eigen_sqrt_diag = np.diag(np.sqrt(eigenvals[0:k]))
    ret = eigenvecs[:, 0:k].dot(eigen_sqrt_diag)
    return ret


def order_groups(X, feature):
    """Assign an order to the values of a categorical feature.

    The function returns an order to the unique values in X[feature] according to
    their similarity based on the other features.
    The distance between two categories is the sum over the distances of each feature.

    Arguments:
    X -- A pandas DataFrame containing all the features to considering in the ordering
    (including the categorical feature to be ordered).
    feature -- String, the name of the column holding the categorical feature to be ordered.
    """

    features = X.columns
    # groups = X[feature].cat.categories.values
    groups = X[feature].unique()
    D_cumu = pd.DataFrame(0, index=groups, columns=groups)
    K = len(groups)
    for j in set(features) - set([feature]):
        D = pd.DataFrame(index=groups, columns=groups)
        # discrete/factor feature j
        # e.g. j = 'color'
        if (X[j].dtypes.name == "category") | (
            (len(X[j].unique()) <= 10) & ("float" not in X[j].dtypes.name)
        ):
            # counts and proportions of each value in j in each group in 'feature'
            cross_counts = pd.crosstab(X[feature], X[j])
            cross_props = cross_counts.div(np.sum(cross_counts, axis=1), axis=0)
            for i in range(K):
                group = groups[i]
                D_values = abs(cross_props - cross_props.loc[group]).sum(axis=1) / 2
                D.loc[group, :] = D_values
                D.loc[:, group] = D_values
        else:
            # continuous feature j
            # e.g. j = 'length'
            # extract the 1/100 quantiles of the feature j
            seq = np.arange(0, 1, 1 / 100)
            q_X_j = X[j].quantile(seq).to_list()
            # get the ecdf (empiricial cumulative distribution function)
            # compute the function from the data points in each group
            X_ecdf = X.groupby(feature)[j].agg(ECDF)
            # apply each of the functions on the quantiles
            # i.e. for each quantile value get the probability that j will take
            # a value less than or equal to this value.
            q_ecdf = X_ecdf.apply(lambda x: x(q_X_j))
            for i in range(K):
                group = groups[i]
                D_values = q_ecdf.apply(lambda x: max(abs(x - q_ecdf[group])))
                D.loc[group, :] = D_values
                D.loc[:, group] = D_values
        D_cumu = D_cumu + D
    # reduce the dimension of the cumulative distance matrix to 1
    D1D = cmds(D_cumu, 1).flatten()
    # order groups based on the values
    order_idx = D1D.argsort()
    groups_ordered = D_cumu.index[D1D.argsort()]
    return pd.Series(range(K), index=groups_ordered)


def quantile_ied(x_vec, q):
    """
    Inverse of empirical distribution function (quantile R type 1).

    More details in
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.mstats.mquantiles.html
    https://stat.ethz.ch/R-manual/R-devel/library/stats/html/quantile.html
    https://en.wikipedia.org/wiki/Quantile

    Arguments:
    x_vec -- A pandas series containing the values to compute the quantile for
    q -- An array of probabilities (values between 0 and 1)
    """

    x_vec = x_vec.sort_values()
    n = len(x_vec) - 1
    m = 0
    j = (n * q + m).astype(int)  # location of the value
    g = n * q + m - j

    gamma = (g != 0).astype(int)
    quant_res = (1 - gamma) * x_vec.shift(1, fill_value=0).iloc[j] + gamma * x_vec.iloc[
        j
    ]
    quant_res.index = q
    # add min at quantile zero and max at quantile one (if needed)
    if 0 in q:
        quant_res.loc[0] = x_vec.min()
    if 1 in q:
        quant_res.loc[1] = x_vec.max()
    return quant_res


def CI_estimate(x_vec, C=0.95):
    """Estimate the size of the confidence interval of a data sample.

    The confidence interval of the given data sample (x_vec) is
    [mean(x_vec) - returned value, mean(x_vec) + returned value].
    """
    alpha = 1 - C
    n = len(x_vec)
    stand_err = x_vec.std() / np.sqrt(n)
    critical_val = 1 - (alpha / 2)
    z_star = stand_err * t.ppf(critical_val, n - 1)
    return z_star



"""Each of the various variable importance methods uses the same code to compute
successively important variables. The only difference between each of these 
methods is the data which is provided to the scoring function. The 
``SelectionStrategy`` handles the process of converting the original training
and scoring data to the form required for each of the individual variables. This
is done by using the current list of important variables to generate a sequence
of triples ``(variable, training_data, scoring_data)``, which will later be 
passed to the scoring function to determine the score for variable.
Below, ``SelectionStrategy`` encapsulates the base functionality which houses the
parameters necessary to produce the generator as well as the default method for
providing only the datasets which are necessary to be evaluated. Each of the
other classes extends this base class to implement a particular variable 
importance method.
If you wish to design your own variable importance method, you will want to
extend the ``SelectionStrategy`` base class in the same way as the other 
strategies.
-----
"""

import numpy as np
import pandas as pd

from .utils import get_data_subset, make_data_from_columns, conditional_permutations

__all__ = [
    "SequentialForwardSelectionStrategy",
    "SequentialBackwardSelectionStrategy",
    "PermutationImportanceSelectionStrategy",
    "SelectionStrategy",
]


class SelectionStrategy(object):
    """The base ``SelectionStrategy`` only provides the tools for storing the
    data and other important information as well as the convenience method for
    iterating over the selection strategies triples lazily."""

    name = "Abstract Selection Strategy"

    def __init__(self, training_data, scoring_data, num_vars, important_vars):
        """Initializes the object by storing the data and keeping track of other
        important information
        :param training_data: (training_inputs, training_outputs)
        :param scoring_data: (scoring_inputs, scoring_outputs)
        :param num_vars: integer for the total number of variables
        :param important_vars: a list of the indices of variables which are already
            considered important
        """
        self.training_data = training_data
        self.scoring_data = scoring_data
        self.num_vars = num_vars
        self.important_vars = important_vars

    def generate_datasets(self, important_variables):
        """Generator which returns triples (variable, training_data, scoring_data)"""
        raise NotImplementedError(
            "Please implement a strategy for generating datasets on class %s"
            % self.name
        )

    def generate_all_datasets(self):
        """By default, loops over all variables not yet considered important"""
        for var in range(self.num_vars):
            if var not in self.important_vars:
                training_data, scoring_data = self.generate_datasets(
                    self.important_vars
                    + [var,]
                )
                yield (training_data, scoring_data, var)

    def __iter__(self):
        return self.generate_all_datasets()


class SequentialForwardSelectionStrategy(SelectionStrategy):
    """Sequential Forward Selection tests all variables which are not yet
    considered important by adding that columns to the other columns which are
    returned. This means that the shape of the training data will be
    ``(num_rows, num_important_vars + 1)``."""

    name = "Sequential Forward Selection"

    def generate_datasets(self, important_variables):
        """Check each of the non-important variables. Dataset is the columns
        which are important
        :returns: (training_data, scoring_data)
        """
        training_inputs, training_outputs = self.training_data
        scoring_inputs, scoring_outputs = self.scoring_data

        columns = important_variables
        # Make a slice of the training inputs
        training_inputs_subset = get_data_subset(training_inputs, None, columns)
        # Make a slice of the scoring inputs
        scoring_inputs_subset = get_data_subset(scoring_inputs, None, columns)
        return (training_inputs_subset, training_outputs), (
            scoring_inputs_subset,
            scoring_outputs,
        )


class SequentialBackwardSelectionStrategy(SelectionStrategy):
    """Sequential Backward Selection tests all variables which are not yet
    considered important by removing that column from the data. This means that
    the shape of the training data will be
    ``(num_rows, num_vars - num_important_vars - 1)``."""

    name = "Sequential Backward Selection"

    def generate_datasets(self, important_variables):
        """Check each of the non-important variables. Dataset is the columns
        which are not important
        :yields: a sequence of (variable being evaluated, columns to include)
        """
        training_inputs, training_outputs = self.training_data
        scoring_inputs, scoring_outputs = self.scoring_data

        columns = [x for x in range(self.num_vars) if x not in important_variables]
        # Make a slice of the training inputs
        training_inputs_subset = get_data_subset(training_inputs, None, columns)
        # Make a slice of the scoring inputs
        scoring_inputs_subset = get_data_subset(scoring_inputs, None, columns)
        return (training_inputs_subset, training_outputs), (
            scoring_inputs_subset,
            scoring_outputs,
        )


class PermutationImportanceSelectionStrategy(SelectionStrategy):
    """Permutation Importance tests all variables which are not yet considered
    important by shuffling that column in addition to the columns of the
    variables which are considered important. The shape of the data will remain
    constant, but at each step, one additional column will be permuted."""

    name = "Permutation Importance"

    def __init__(
        self,
        training_data,
        scoring_data,
        num_vars,
        important_vars,
        random_state,
        **kwargs
    ):
        """Initializes the object by storing the data and keeping track of other
        important information
        :param training_data: (training_inputs, training_outputs)
        :param scoring_data: (scoring_inputs, scoring_outputs)
        :param num_vars: integer for the total number of variables
        :param important_vars: a list of the indices of variables which are
            already considered important
        """
        super(PermutationImportanceSelectionStrategy, self).__init__(
            training_data, scoring_data, num_vars, important_vars
        )

        # Also initialize the "shuffled data"
        scoring_inputs, __ = self.scoring_data
        indices = random_state.permutation(len(scoring_inputs))
        
        # With each iteration of the algorithm, the indices 
        # are shuffled once and identically for each feature. 
        # Thus, when multiple features are permuted they are 
        # jointly permuted (i.e., without destroying the 
        # dependencies of the features within the group).
        # However, how the features are jointly permuted 
        # changes from iteration to iteration to limit
        # bias due to a poor permutation. 
        
        self.shuffled_scoring_inputs = get_data_subset(
            scoring_inputs, indices
        )  # This copies
        # keep track of the initial index (assuming this is pandas data)
        self.original_index = (
            scoring_inputs.index if isinstance(scoring_inputs, pd.DataFrame) else None
        )

    def generate_datasets(self, important_variables):
        """Check each of the non-important variables. Dataset has columns which
        are important shuffled
        :returns: (training_data, scoring_data)
        """
        scoring_inputs, scoring_outputs = self.scoring_data
        complete_scoring_inputs = make_data_from_columns(
            [
                get_data_subset(
                    self.shuffled_scoring_inputs
                    if i in important_variables
                    else scoring_inputs,
                    None,
                    [i],
                )
                for i in range(self.num_vars)
            ],
            index=self.original_index,
        )

        return self.training_data, (complete_scoring_inputs, scoring_outputs)


class ConditionalPermutationImportanceSelectionStrategy(SelectionStrategy):
    """Conditional Permutation Importance tests all variables which are not yet considered
    important by performing conditional permutation on that column in addition to the columns of the
    variables which are considered important. The shape of the data will remain
    constant, but at each step, one additional column will be permuted."""

    name = "Conditional Permutation Importance"

    def __init__(
        self,
        training_data,
        scoring_data,
        num_vars,
        important_vars,
        random_state,
        **kwargs
    ):
        """Initializes the object by storing the data and keeping track of other
        important information
        :param training_data: (training_inputs, training_outputs)
        :param scoring_data: (scoring_inputs, scoring_outputs)
        :param num_vars: integer for the total number of variables
        :param important_vars: a list of the indices of variables which are
            already considered important
        """
        super(ConditionalPermutationImportanceSelectionStrategy, self).__init__(
            training_data, scoring_data, num_vars, important_vars
        )
        n_bins = kwargs.get("n_bins", 50)

        # Also initialize the "shuffled data"
        scoring_inputs, __ = self.scoring_data
        self.shuffled_scoring_inputs = conditional_permutations(
            scoring_inputs, n_bins, random_state
        )  # This copies
        # keep track of the initial index (assuming this is pandas data)
        self.original_index = (
            scoring_inputs.index if isinstance(scoring_inputs, pd.DataFrame) else None
        )

    def generate_datasets(self, important_variables):
        """Check each of the non-important variables. Dataset has columns which
        are important shuffled
        :returns: (training_data, scoring_data)
        """
        scoring_inputs, scoring_outputs = self.scoring_data
        # If a feature has been deemed important it remains shuffled
        complete_scoring_inputs = make_data_from_columns(
            [
                get_data_subset(
                    self.shuffled_scoring_inputs
                    if i in important_variables
                    else scoring_inputs,
                    None,
                    [i],
                )
                for i in range(self.num_vars)
            ],
            index=self.original_index,
        )

        return self.training_data, (complete_scoring_inputs, scoring_outputs)


class ForwardPermutationImportanceSelectionStrategy(SelectionStrategy):
    """Forward Permutation Importance permutes all variables and then tests
    all variables which are not yet considered."""

    name = "Forward Permutation Importance"

    def __init__(
        self,
        training_data,
        scoring_data,
        num_vars,
        important_vars,
        random_state,
        **kwargs
    ):
        """Initializes the object by storing the data and keeping track of other
        important information
        :param training_data: (training_inputs, training_outputs)
        :param scoring_data: (scoring_inputs, scoring_outputs)
        :param num_vars: integer for the total number of variables
        :param important_vars: a list of the indices of variables which are
            already considered important
        """
        super(ForwardPermutationImportanceSelectionStrategy, self).__init__(
            training_data, scoring_data, num_vars, important_vars
        )

         # With each iteration of the algorithm, the indices 
        # are shuffled once and identically for each feature. 
        # Thus, when multiple features are permuted they are 
        # jointly permuted (i.e., without destroying the 
        # dependencies of the features within the group).
        # However, how the features are jointly permuted 
        # changes from iteration to iteration to limit
        # bias due to a poor permutation. 
        
        
        # Also initialize the "shuffled data"
        scoring_inputs, __ = self.scoring_data
        indices = random_state.permutation(len(scoring_inputs))
        self.shuffled_scoring_inputs = get_data_subset(
            scoring_inputs, indices
        )  # This copies
        # keep track of the initial index (assuming this is pandas data)
        self.original_index = (
            scoring_inputs.index if isinstance(scoring_inputs, pd.DataFrame) else None
        )
        
    def generate_datasets(self, important_variables):
        """Check each of the non-important variables. Dataset has columns which
        are non-important variables are shuffled
        :returns: (training_data, scoring_data)
        """
        scoring_inputs, scoring_outputs = self.scoring_data
        # If a feature has been deemed important it remains shuffled
        complete_scoring_inputs = make_data_from_columns(
            [
                get_data_subset(
                    scoring_inputs
                    if i in important_variables
                    else self.shuffled_scoring_inputs,
                    columns = [i],
                )
                for i in range(self.num_vars)
            ],
            index=self.original_index,
        )

        return self.training_data, (complete_scoring_inputs, scoring_outputs)



import shap
import traceback
from .tree_interpreter import TreeInterpreter
import pandas as pd
import numpy as np

from lime.lime_tabular import LimeTabularExplainer

from ..common.attributes import Attributes
from ..common.importance_utils import retrieve_important_vars
from ..common.utils import to_dataframe, determine_feature_dtype
from ..common.metrics import brier_skill_score


from ..common.contrib_utils import (
    get_indices_based_on_performance,
    avg_and_sort_contributions,
)


list_of_acceptable_tree_estimators = [
    "RandomForestClassifier",
    "RandomForestRegressor",
    "DecisionTreeClassifier",
    "ExtraTreesClassifier",
    "ExtraTreesRegressor",
]


class LocalExplainer(Attributes):

    """
    ExplainToolkit inherits functionality from LocalExplainer and is not meant to be
    instantiated by the end-user.


    LocalExplainer incorporates important methods for explaining local estimator behavior
    for select data instances. The calculations are primarily based on SHAP (source), but also
    includes treeinterpreter (source) for random forests and for other select tree-based methods in
    scikit-learn (see list_of_acceptable_tree_estimators).

    Attributes
    --------------
    estimators : object, list of objects
        A fitted estimator object or list thereof implementing `predict` or
        `predict_proba`.
        Multioutput-multiclass classifiers are not supported.

    estimator_names : string, list
        Names of the estimators (for internal and plotting purposes)

    X : {array-like or dataframe} of shape (n_samples, n_features)
        Training or validation data used to compute the IML methods.
        If ndnumpy array, make sure to specify the feature names

    y : {list or numpy.array} of shape = (n_samples,)
        y values.

    estimator_output : "raw" or "probability"
        What output of the estimator should be evaluated. default is None. If None,
        ExplainToolkit will determine internally what the estimator output is.

    feature_names : array-like of shape (n_features,), dtype=str, default=None
        Name of each feature; `feature_names[i]` holds the name of the feature
        with index `i`.
        By default, the name of the feature corresponds to their numerical
        index for NumPy array and their column name for pandas dataframe.
        Feature names are only required if X is an ndnumpy.array, it will be
        converted to a pandas.DataFrame internally.


    Reference:
        SHAP
        treeinterpreter
    """

    def __init__(
        self,
        estimators,
        estimator_names,
        X,
        y,
        estimator_output,
        feature_names=None,
        checked_attributes=False,
    ):
        # These functions come from the inherited Attributes class
        if not checked_attributes:
            self.set_estimator_attribute(estimators, estimator_names)
            self.set_X_attribute(X, feature_names)
            self.set_y_attribute(y)
            self.set_estimator_output(estimator_output, estimators)
        else:
            self.estimators = estimators
            self.estimator_names = estimator_names
            self.X = X
            self.y = y
            self.feature_names = list(X.columns)
            self.estimator_output = estimator_output

    def _get_local_prediction(
        self,
        method="shap",
        performance_based=True,
        n_samples=100,
        shap_kwargs=None,
        lime_kws=None, 
    ):
        """
        Explain individual predictions using SHAP (SHapley Additive exPlanations;
        https://github.com/slundberg/shap) or treeinterpreter
        (https://blog.datadive.net/random-forest-interpretation-with-scikit-learn/)

        Parameters
        ------------
        method : 'treeinterpreter', 'shap', or 'lime'
            Contributions can be computed using treeinterpreter for tree-based estimators
            or using SHAP for both tree- and non-tree-based estimators
        background_dataset : array (n_samples, n_features)
            A representative (often a K-means or random sample) subset of the
            data used to train the ML estimator. Used for the background dataset
            to compute the expected values for the SHAP calculations.
            Only required for non-tree based estimators.
        performance_based : boolean (default=False)
            If True, will average feature contributions over the best and worst
            performing of the given X. The number of X to average over
            is given by n_samples
        n_samples : interger (default=100)
            Number of samples to compute average over if performance_based = True

        shap_kwargs : dict
            Arguments passed to the shap.Explainer object. See
            https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer
            for details. The main two arguments supported in PyMint is the masker and
            algorithm options. By default, the masker option uses
            masker = shap.maskers.Partition(X, max_samples=100, clustering="correlation") for
            hierarchical clustering by correlations. You can also provide a background dataset
            e.g., background_dataset = shap.sample(X, 100).reset_index(drop=True). The algorithm
            option is set to "auto" by default.

            - masker
            - algorithm


        Returns
        ---------

        results_ds : pandas.DataFrame

        """
        if method not in ["tree_interpreter", "shap", "lime"]:
            raise ValueError(
                """
                             Declared method is not 'tree_interpreter','shap', or 'lime'. 
                             Check for spelling mistake or syntax error!
                             """
            )
        else:
            self.method = method

        # will be returned; a list of pandas dataframes, one for each performance dict key
        contributions_dict = {
            estimator_name: {} for estimator_name in self.estimator_names
        }
        feature_values_dict = {
            estimator_name: {} for estimator_name in self.estimator_names
        }

        for estimator_name, estimator in self.estimators.items():
            # create entry for current estimator
            # self.contributions_dict[estimator_name] = {}
            if performance_based:
                ### print('Computing performance-based contributions...')
                performance_dict = get_indices_based_on_performance(
                    estimator,
                    X=self.X,
                    y=self.y,
                    n_samples=n_samples,
                    estimator_output=self.estimator_output,
                )

                for key, indices in performance_dict.items():
                    cont_dict = self._get_feature_contributions(
                        estimator=estimator,
                        X=self.X.iloc[indices, :],
                        shap_kwargs=shap_kwargs,
                    )

                    contributions_dict[estimator_name][key] = cont_dict
                    feature_values_dict[estimator_name][key] = self.X.iloc[indices, :]

            else:
                cont_dict = self._get_feature_contributions(
                    estimator=estimator,
                    X=self.X,
                    shap_kwargs=shap_kwargs,
                    lime_kws=lime_kws, 
                )

                contributions_dict[estimator_name]["non_performance"] = cont_dict
                feature_values_dict[estimator_name]["non_performance"] = self.X

            # average out the contributions and sort based on contribution
            avg_contrib_dict, avg_feature_val_dict = avg_and_sort_contributions(
                contributions_dict[estimator_name], feature_values_dict[estimator_name]
            )

            contributions_dict[estimator_name] = avg_contrib_dict
            feature_values_dict[estimator_name] = avg_feature_val_dict

        results = (contributions_dict, feature_values_dict)

        results_df = to_dataframe(results, self.estimator_names, self.feature_names)

        return results_df

    def _get_shap_values(
        self,
        estimator,
        X,
        shap_kwargs=None,
    ):
        """
        FOR INTERNAL PURPOSES ONLY.

        """
        masker = shap_kwargs.get("masker", None)
        algorithm = shap_kwargs.get("algorithm", "auto")

        if masker is None:
            raise ValueError(
                """masker in shap_kwargs is None. 
                             This will cause issues with SHAP. We recommend starting with
                             shap_kwargs = {'masker' = shap.maskers.Partition(X, max_samples=100, clustering="correlation")}
                             where X is the original dataset and not the examples SHAP is being computed for. 
                             """
            )

        if self.estimator_output == "probability":
            model = estimator.predict_proba
        else:
            model = estimator.predict

        explainer = shap.Explainer(
            model=model,
            masker=masker,
            algorithm=algorithm,
        )

        shap_results = explainer(X)

        if self.estimator_output == "probability":
            shap_results = shap_results[..., 1]

        contributions = shap_results.values
        bias = shap_results.base_values
        
        return contributions, bias

    def _get_ti_values(self, estimator, X):
        """
        FOR INTERNAL PURPOSES ONLY.
        """
        # check to make sure estimator is of type Tree
        if type(estimator).__name__ not in list_of_acceptable_tree_estimators:
            raise TypeError(
                f""" Unfortunately, tree interpreter does not work on this type of estimator :
                                {type(estimator).__name__}
                            """
            )

        ti = TreeInterpreter(estimator, X)

        prediction, bias, contributions = ti.predict()

        if self.estimator_output == "probability":
            contributions = contributions[:, :, 1]
            bias = bias[:, 1]  # bias is all the same values for first index
        else:
            pass

        return contributions, bias

    
    def _get_lime_values(self, estimator, X, lime_kws):
        """
        Compute the Local Interpretable Model-Agnostic Explanations
        """
        if lime_kws is None:
            raise KeyError('lime_kws is None, but lime_kws must contain training_data!')
        
        lime_kws['feature_names'] = X.columns
        
        # Determine categorical features
        if lime_kws.get('categorical_names', None) is None and lime_kws.get('categorical_features', None) is None:
            features, cat_features = determine_feature_dtype(X, X.columns)
            lime_kws['categorical_names'] = cat_features
        
        
        # Determine whether its a classification or regression task.
        mode = 'regression' 
        if lime_kws.get('mode', None) is None:
            mode = 'classification' if hasattr(estimator, 'predict_proba') else 'regression'
            lime_kws['mode'] = mode
        
        # Set the random state for reproducible results. 
        if lime_kws.get('random_state', None) is None:
            lime_kws['random_state'] = 123 
        
        
        explainer = LimeTabularExplainer(**lime_kws)
        
        if mode == 'classification' and hasattr(estimator, 'predict_proba'):
            predict_fn = estimator.predict_proba 
        else:
            predict_fn = estimator.predict 
        
        # Using X.values[0] rather than X.values since it requires 
        # a (n_features) rather than (1,n_features).
        explanation = explainer.explain_instance(X.values[0], predict_fn, num_features=X.shape[1],)

        sorted_exp = sorted(explanation.local_exp[1], key=lambda x: x[0])
        contributions = np.array([[val[1] for val in sorted_exp]])
        bias = explanation.intercept[1] 
    
        return contributions, bias
        
    
    def _get_feature_contributions(self, estimator, X, shap_kwargs=None, lime_kws=None):
        """
        FOR INTERNAL PURPOSES ONLY.

        Compute the feature contributions either using treeinterpreter or SHAP.

        Parameters
        ------------
        estimator : callable
        X : pandas.DataFrame of shape (n_samples, n_features)

        """
        if self.method == "shap":
            contributions, bias = self._get_shap_values(estimator, X, shap_kwargs)
        elif self.method == "tree_interpreter":
            contributions, bias = self._get_ti_values(estimator, X)
        elif self.method == 'lime':
            contributions, bias = self._get_lime_values(estimator, X, lime_kws)

        n_samples = len(X)
        columns = self.feature_names + ["Bias"]

        if self.estimator_output == "probability":
            contributions *= 100.0
            bias *= 100.

        # A single example.
        if isinstance(bias, float):
            bias = [bias]

        bias_reshaped = np.reshape(bias, (len(bias), 1))
        data = np.concatenate((contributions, bias_reshaped), axis=1)

        return pd.DataFrame(data, columns=columns)



import numpy as np
import pandas as pd
import xarray as xr
from math import sqrt
from scipy.spatial import cKDTree
from scipy.interpolate import interp1d
import itertools
from functools import reduce
from operator import add
import traceback
from copy import copy
from inspect import currentframe, getframeinfo
from pandas.core.common import SettingWithCopyError
from joblib import delayed, Parallel


from sklearn.linear_model import LinearRegression
from sklearn import cluster
from sklearn.metrics import (
    roc_auc_score,
    roc_curve,
    average_precision_score,
    mean_squared_error,
    r2_score,
)

from ..common.utils import (
    compute_bootstrap_indices,
    merge_nested_dict,
    merge_dict,
    is_str,
    is_valid_feature,
    is_regressor,
    is_classifier,
    cartesian,
    to_xarray,
    order_groups,
    determine_feature_dtype,
    check_is_permuted,
)

from ..common.importance_utils import to_skexplain_importance, all_permuted_score
from ..common.metrics import brier_skill_score, norm_aupdc

from ..common.multiprocessing_utils import run_parallel, to_iterator
from ..common.attributes import Attributes

from .PermutationImportance import sklearn_permutation_importance
from .PermutationImportance.utils import bootstrap_generator


class GlobalExplainer(Attributes):

    """
    ExplainToolkit inherits functionality from GlobalExplainer and is not meant to be
    instantiated by the end-user.

    GlobalExplainer incorporates important methods for explaining global estimator behavior
    across all data instances. These include permutation importance and
    partial dependence (Friedman 2001) / accumulated local effects (Apley and Zhy et al. 2016)
    which produce feature ranking and expected relationship between a feature
    with the predict outcome, respectively.

    The permutation importance computations rely on the
    PermutationImportance python package (Jergensen 2019) with slight modification for
    use in py-mint.

    Parts of the partial dependence code were based on
    the computations in sklearn.inspection.partial_dependence (Pedregosa et al. 2011).

    Parts of the accumulated local effects
    are based on ALEPython (Jumelle 2020) and PyALE.


    PartialDependence is a class for computing first- and second-order
    partial dependence (PD; Friedman (2001). Parts of the code were based on
    the computations in sklearn.inspection.partial_dependence (Pedregosa et al. 2011).
    Currently, the package handles regression and binary classification.

    Attributes
    --------------
    estimators : object, list of objects
        A fitted estimator object or list thereof implementing `predict` or
        `predict_proba`.
        Multioutput-multiclass classifiers are not supported.

    estimator_names : string, list
        Names of the estimators (for internal and plotting purposes)

    X : {array-like or dataframe} of shape (n_samples, n_features)
        Training or validation data used to compute the IML methods.
        If ndnumpy array, make sure to specify the feature names

    y : {list or numpy.array} of shape = (n_samples,)
        y values.

    estimator_output : "raw" or "probability"
        What output of the estimator should be evaluated. default is None. If None,
        ExplainToolkit will determine internally what the estimator output is.

    feature_names : array-like of shape (n_features,), dtype=str, default=None
        Name of each feature; `feature_names[i]` holds the name of the feature
        with index `i`.
        By default, the name of the feature corresponds to their numerical
        index for NumPy array and their column name for pandas dataframe.
        Feature names are only required if X is an ndnumpy.array, it will be
        converted to a pandas.DataFrame internally.

    Reference
    ------------
        Friedman, J. H., 2001: GREEDY FUNCTION APPROXIMATION: A GRADIENT BOOSTING MACHINE.
        Ann Statistics, 29, 1189–1232, https://doi.org/10.1214/aos/1013203451.

        Jumelle, M., 2020: ALEPython.
        Github Python software library https://github.com/blent-ai/ALEPython.

        Jergensen, G., 2019: PermutationImportance.
        Github Python software library https://github.com/gelijergensen/PermutationImportance.

        Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011.
    """

    def __init__(
        self,
        estimators,
        estimator_names,
        X,
        y=None,
        estimator_output=None,
        feature_names=None,
        checked_attributes=False,
    ):
        # These functions come from the inherited Attributes class
        if not checked_attributes:
            self.set_estimator_attribute(estimators, estimator_names)
            self.set_X_attribute(X, feature_names)
            self.set_y_attribute(y)
        else:
            self.estimators = estimators
            self.estimator_names = estimator_names
            self.X = X
            self.y = y
            self.feature_names = list(X.columns)

        self.estimator_output = estimator_output

    def _to_scorer(self, evaluation_fn):
        """
        FOR INTERNAL PURPOSES ONLY.
        Converts a string to an evaluation function.
        """
        available_scores = ["auc", "auprc", "bss", "mse", "norm_aupdc"]

        if evaluation_fn == "auc":
            evaluation_fn = roc_auc_score
            scoring_strategy = "argmin_of_mean"
        elif evaluation_fn == "auprc":
            evaluation_fn = average_precision_score
            scoring_strategy = "argmin_of_mean"
        elif evaluation_fn == "norm_aupdc":
            evaluation_fn = norm_aupdc
            scoring_strategy = "argmin_of_mean"
        elif evaluation_fn == "bss":
            evaluation_fn = brier_skill_score
            scoring_strategy = "argmin_of_mean"
        elif evaluation_fn == "mse":
            evaluation_fn = mean_squared_error
            scoring_strategy = "argmax_of_mean"
        else:
            raise ValueError(
                f"evaluation_fn is not set! Available options are {available_scores}"
            )
        return evaluation_fn, scoring_strategy

    def calc_permutation_importance(
        self,
        n_vars=5,
        evaluation_fn="auprc",
        subsample=1.0,
        n_jobs=1,
        n_permute=1,
        scoring_strategy=None,
        direction="backward",
        verbose=False,
        return_iterations=True,
        random_seed=1,
    ):

        """
        Performs single-pass and/or multi-pass permutation importance using the PermutationImportance
        package.

        See calc_permutation_importance in IntepretToolkit for documentation.

        """
        is_str = False
        if isinstance(evaluation_fn, str):
            evaluation_fn = evaluation_fn.lower()
            is_str = True

        if not isinstance(evaluation_fn, str) and scoring_strategy is None:
            raise ValueError(
                """ 
                The scoring_strategy argument is None! If you are using a non-default evaluation_fn 
                then scoring_strategy must be set! If the metric is positively-oriented (a higher value is better), 
                then set scoring_strategy = "argmin_of_mean" and if it is negatively-oriented-
                (a lower value is better), then set scoring_strategy = "argmax_of_mean"
                """
            )

        if isinstance(evaluation_fn, str):
            evaluation_fn, scoring_strategy = self._to_scorer(evaluation_fn)

        # Convert the scoring strategies to the appropriate 
        # style for PermutationImportance. 
        if scoring_strategy == 'maximize':
            scoring_strategy == 'argmax_of_mean'
        elif scoring_strategy == 'minimize':
            scoring_strategy == 'argmin_of_mean'

        if is_str:
            if direction == "forward":
                if "max" in scoring_strategy:
                    scoring_strategy = scoring_strategy.replace("max", "min")
                else:
                    scoring_strategy = scoring_strategy.replace("min", "max")        
                    
        y = pd.DataFrame(data=self.y, columns=["Test"])

        pi_dict = {}

        # loop over each estimator
        for estimator_name, estimator in self.estimators.items():
            pi_result = sklearn_permutation_importance(
                model=estimator,
                scoring_data=(self.X.values, y.values),
                evaluation_fn=evaluation_fn,
                variable_names=self.feature_names,
                scoring_strategy=scoring_strategy,
                subsample=subsample,
                nimportant_vars=n_vars,
                njobs=n_jobs,
                n_permute=n_permute,
                verbose=verbose,
                direction=direction,
                random_seed=random_seed,
            )
            if direction == 'forward':
                scores = all_permuted_score(estimator, 
                                            self.X.values, 
                                            y.values, 
                                            evaluation_fn=evaluation_fn, 
                                            n_permute=n_permute, 
                                            subsample=subsample, random_seed=random_seed)
                
                pi_result.permuted_score = scores
            
            pi_dict[estimator_name] = pi_result

            del pi_result

        data = {}
        for estimator_name in self.estimator_names:
            for func in ["retrieve_multipass", "retrieve_singlepass"]:
                adict = getattr(pi_dict[estimator_name], func)()
                features = np.array(list(adict.keys()))
                rankings = np.argsort([adict[f][0] for f in features])
                top_features = features[rankings]
                scores = np.array([adict[f][1] for f in top_features])
                pass_method = func.split("_")[1]

                data[f"{pass_method}_rankings__{estimator_name}"] = (
                    [f"n_vars_{pass_method}"],
                    top_features,
                )
                data[f"{pass_method}_scores__{estimator_name}"] = (
                    [f"n_vars_{pass_method}", "n_permute"],
                    scores,
                )
                
            data[f"original_score__{estimator_name}"] = (
                ["n_permute"],
                pi_dict[estimator_name].original_score,
            )
            
            if direction == 'forward':
                data[f"all_permuted_score__{estimator_name}"] = (
                 ["n_permute"],
                 pi_dict[estimator_name].permuted_score,
            )
            

        if return_iterations:
            for estimator_name in self.estimator_names:
                temp_results = getattr(
                    pi_dict[estimator_name], "retrieve_all_iterations"
                )()
                temp_scores = []
                temp_features = []
                for adict in temp_results:
                    features = np.array(list(adict.keys()))
                    rankings = np.argsort([adict[f][0] for f in features])
                    top_features = features[rankings]
                    scores = np.array([adict[f][1] for f in top_features])
                    pass_method = func.split("_")[1]
                    # Just retrieve the second most important feature.
                    temp_scores.append(scores[1])
                    temp_features.append(top_features[1])

                data[f"second_place_scores__{estimator_name}"] = (
                    [f"n_vars_second_place", "n_permute"],
                    temp_scores,
                )
                data[f"second_place_rankings__{estimator_name}"] = (
                    [f"n_vars_second_place"],
                    temp_features,
                )

        results_ds = to_xarray(data)

        # Determine the orientation of the loss metric used.
        # If 'max' in strategy, then 'negative' and
        # if 'min', then 'positive'
        #orientation = "positive" if "min" in scoring_strategy else "negative"

        return results_ds, scoring_strategy

    def _run_interpret_curves(
        self,
        method,
        features=None,
        n_bins=25,
        n_jobs=1,
        subsample=1.0,
        n_bootstrap=1,
        feature_encoder=None,
        random_seed=42,
    ):

        """
        Runs the interpretation curve (partial dependence, accumulated local effects,
        or individual conditional expectations.) calculations.
        Includes assessing whether the calculation is 1D or 2D and handling
        initializing the parallelization, subsampling data, and/or using bootstraping
        to compute confidence intervals.

        Returns a nested dictionary with all neccesary inputs for plotting.

        Parameters
        ---------
        method : 'pd' , 'ale', or 'ice'
                determines whether to compute partial dependence ('pd'),
                accumulated local effects ('ale'), or individual conditional expectations ('ice')

        features: string, 2-tuple of strings, list of strings, or lists of 2-tuple strings
                feature names to compute for. If 2-tuple, it will compute
                the second-order results.

        n_bins : int
            Number of evenly-spaced bins to compute PD/ALE over.

        n_jobs : int or float
            if int, the number of processors to use for parallelization
            if float, percentage of total processors to use for parallelization

        subsample : float or integer
            if value between 0-1 interpreted as fraction of total X to use
                    if value > 1, interpreted as the number of X to randomly sample
                        from the original dataset.

        n_bootstrap: integer
            Number of bootstrap iterations to perform. Defaults to 1 (no bootstrapping).

        Returns
        ---------

        results_ds : xarray.Dataset

        """
        self.random_seed = random_seed
        # Check if features is a string
        if is_str(features) or isinstance(features, tuple):
            features = [features]

        if not isinstance(features[0], tuple):
            features, cat_features = determine_feature_dtype(self.X, features)
        else:
            cat_features = []

        results = []
        cat_results = []
        if len(features) > 0:
            if method == "ale":
                # check first element of feature and see if the type is a tuple; assume second-order calculations
                if isinstance(features[0], tuple):
                    func = self.compute_second_order_ale
                else:
                    func = self.compute_first_order_ale
            elif method == "pd":
                func = self.compute_partial_dependence
            elif method == "ice":
                func = self.compute_individual_cond_expect

            args_iterator = to_iterator(
                self.estimator_names,
                features,
                [n_bins],
                [subsample],
                [n_bootstrap],
            )

            total = len(features) * len(self.estimator_names)
            results = run_parallel(
                func=func,
                args_iterator=args_iterator,
                kwargs={},
                nprocs_to_use=n_jobs,
                total=total,
            )

        if len(cat_features) > 0:
            if method == "ale":
                func = self.compute_first_order_ale_cat
            elif method == "pd":
                func = self.compute_partial_dependence
            elif method == "ice":
                func = self.compute_individual_cond_expect

            args_iterator = to_iterator(
                self.estimator_names,
                cat_features,
                [subsample],
                [n_bootstrap],
                [feature_encoder],
            )

            total = len(cat_features) * len(self.estimator_names)
            cat_results = run_parallel(
                func=func,
                args_iterator=args_iterator,
                kwargs={},
                nprocs_to_use=n_jobs,
                total=total,
            )

        results = cat_results + results

        results = merge_dict(results)
        results_ds = to_xarray(results)

        return results_ds

    def _store_results(
        self,
        method,
        estimator_name,
        features,
        ydata,
        xdata,
        hist_data,
        ice_X=None,
        categorical=False,
    ):
        """
        FOR INTERNAL PURPOSES ONLY.

        Store the results of the ALE/PD/ICE calculations into a dict,
        which is converted to an xarray.Dataset

        """
        results = {}

        feature1 = f"{features[0]}" if isinstance(features, tuple) else features
        feature2 = f"__{features[1]}" if isinstance(features, tuple) else ""

        y_shape = ["n_bootstrap", f"n_bins__{feature1}", f"n_bins{feature2}"]
        if feature2 == "":
            y_shape = y_shape[:-1]

        hist_data1 = hist_data[0]

        xdata2 = None
        if method == "pd" or method == "ice" or (method == "ale" and categorical):
            xdata1 = xdata[0]
            if np.shape(xdata)[0] > 1:
                xdata2 = xdata[1]
                hist_data2 = hist_data[1]
        else:
            if feature2 != "":
                xdata1 = 0.5 * (xdata[0][1:] + xdata[0][:-1])
                xdata2 = 0.5 * (xdata[1][1:] + xdata[1][:-1])
            else:
                xdata1 = 0.5 * (xdata[1:] + xdata[:-1])

            if feature2 != "":
                hist_data2 = hist_data[1]

        results[f"{feature1}{feature2}__{estimator_name}__{method}"] = (y_shape, ydata)
        results[f"{feature1}__bin_values"] = ([f"n_bins__{feature1}"], xdata1)
        results[f"{feature1}"] = (["n_X"], hist_data1)
        if feature2 != "":
            results[f"{feature2[2:]}__bin_values"] = ([f"n_bins{feature2[2:]}"], xdata2)
            results[f"{feature2[2:]}"] = (["n_X"], hist_data2)

        if ice_X is not None:
            results["X_sampled"] = (["n_samples", "n_features"], ice_X)
            results["features"] = (["n_features"], ice_X.columns)

        return results

    def compute_individual_cond_expect(
        self,
        estimator_name,
        features,
        n_bins=30,
        subsample=1.0,
        n_bootstrap=1,
    ):
        """
        Compute the Individual Conditional Expectations (see https://christophm.github.io/interpretable-ml-book/ice.html)

        Parameters
        -----------

        estimator_name : string
            Name of the estimator; used as a dict keys

        features : string or 2-tuple of strings
            Feature name or pair of feature names to compute ICE for.

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.


        Returns
        --------

        results : dict

        """
        random_state = np.random.RandomState(self.random_seed)

        # Retrieve the estimator object from the estimators dict attribute
        estimator = self.estimators[estimator_name]

        # Check if features is a string
        if is_str(features):
            features = [features]

        # Check if feature is valid
        is_valid_feature(features, self.feature_names)

        if float(subsample) != 1.0:
            n_X = len(self.X)
            size = int(n_X * subsample) if subsample <= 1.0 else subsample
            idxs = random_state.choice(n_X, size=size, replace=False)
            X = self.X.iloc[idxs, :]
            X.reset_index(drop=True, inplace=True)
        else:
            X = self.X.copy()

        # Extract the values for the features
        feature_values = [X[f].to_numpy() for f in features]

        # Create a grid of values
        grid = [np.linspace(np.amin(f), np.amax(f), n_bins) for f in feature_values]

        if self.estimator_output == "probability":
            prediction_method = estimator.predict_proba
        elif self.estimator_output == "raw":
            prediction_method = estimator.predict

        ice_values = []
        for value_set in cartesian(grid):
            X_temp = X.copy()
            X_temp.loc[:, features[0]] = value_set[0]
            ice_values.append(prediction_method(X_temp.values))

        ice_values = np.array(ice_values).T
        if self.estimator_output == "probability":
            # Binary classification, shape is (2, n_points).
            # we output the effect of **positive** class
            # and convert to percentages
            ice_values = ice_values[1]

        # center the ICE plots
        ice_values -= np.mean(ice_values, axis=1).reshape(len(ice_values), 1)

        if len(features) > 1:
            ice_values = ice_values.reshape(n_bootstrap, n_bins, n_bins)
        else:
            features = features[0]

        results = self._store_results(
            method="ice",
            estimator_name=estimator_name,
            features=features,
            ydata=ice_values,
            xdata=grid,
            hist_data=feature_values,
            ice_X=X,
        )

        return results

    def compute_partial_dependence(
        self, estimator_name, features, n_bins=30, subsample=1.0, n_bootstrap=1
    ):

        """
        Calculate the centered partial dependence.

        # Friedman, J., 2001: Greedy function approximation: a gradient boosting machine.
        Annals of Statistics, 29 (5), 1189–1232.
        ##########################################################################
        # Partial dependence plots fix a value for one or more predictors
        # for X, passing these new data through a trained estimator,
        # and then averaging the resulting predictions. After repeating this process
        # for a range of values of X*, regions of non-zero slope indicates that
        # where the ML estimator is sensitive to X* (McGovern et al. 2019). Only disadvantage is
        # that PDP do not account for non-linear interactions between X and the other predictors.
        #########################################################################

        Parameters
        --------------
        estimator_name : string
            Name of the estimator; used as a dict keys

        features : string or 2-tuple of strings
            Feature name or pair of feature names to compute ICE for.

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.

        Returns
        ------------------
            results : dict
        """
        # Retrieve the estimator object from the estimators dict attribute
        estimator = self.estimators[estimator_name]

        # Check if features is a string
        if is_str(features):
            features = [features]

        # Check if feature is valid
        is_valid_feature(features, self.feature_names)

        # Extract the values for the features
        full_feature_values = [self.X[f].to_numpy() for f in features]

        # Create a grid of values
        grid = [
            np.linspace(np.amin(f), np.amax(f), n_bins) for f in full_feature_values
        ]

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X,
                subsample=subsample,
                n_bootstrap=n_bootstrap,
                seed=self.random_seed,
            )
        else:
            bootstrap_indices = [self.X.index.to_list()]

        if self.estimator_output == "probability":
            prediction_method = estimator.predict_proba
        elif self.estimator_output == "raw":
            prediction_method = estimator.predict

        pd_values = []

        # for each bootstrap set
        for k, idx in enumerate(bootstrap_indices):
            # get samples
            X = self.X.iloc[idx, :].reset_index(drop=True)
            feature_values = [X[f].to_numpy() for f in features]
            averaged_predictions = []
            # for each value, set all indices to the value,
            # make prediction, store mean prediction
            for value_set in cartesian(grid):
                X_temp = X.copy()
                for i, feature in enumerate(features):
                    X_temp.loc[:, feature] = value_set[i]
                    predictions = prediction_method(X_temp.values)

                # average over samples
                averaged_predictions.append(np.mean(predictions, axis=0))

            averaged_predictions = np.array(averaged_predictions).T

            if self.estimator_output == "probability":
                # Binary classification, shape is (2, n_points).
                # we output the effect of **positive** class
                # and convert to percentages
                averaged_predictions = averaged_predictions[1]

            # Center the predictions
            averaged_predictions -= np.mean(averaged_predictions)

            pd_values.append(averaged_predictions)

        # Reshape the pd_values for higher-order effects
        pd_values = np.array(pd_values)
        if len(features) > 1:
            pd_values = pd_values.reshape([n_bootstrap] + [n_bins] * len(features))
        else:
            features = features[0]

        results = self._store_results(
            method="pd",
            estimator_name=estimator_name,
            features=features,
            ydata=pd_values,
            xdata=grid,
            hist_data=feature_values,
        )

        return results

    def compute_first_order_ale(
        self, estimator_name, feature, n_bins=30, subsample=1.0, n_bootstrap=1
    ):
        """
        Computes first-order ALE function on single continuous feature data.

        Script is largely the _first_order_ale_quant from
        https://github.com/blent-ai/ALEPython/ with small modifications.

        Parameters
        ----------
        estimator_name : string
            Name of the estimator; used as a dict keys

        features : string or 2-tuple of strings
            Feature name or pair of feature names to compute ICE for.

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.


        Returns
        ----------
            results : nested dictionary

        """
        estimator = self.estimators[estimator_name]
        # check to make sure feature is valid
        if feature not in self.feature_names:
            raise KeyError(f"Feature {feature} is not a valid feature.")

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X,
                subsample=subsample,
                n_bootstrap=n_bootstrap,
                seed=self.random_seed,
            )
        else:
            bootstrap_indices = [self.X.index.to_list()]

        # Using the original, unaltered feature values
        # calculate the bin edges to be used in the bootstrapping.
        original_feature_values = self.X[feature].values

        if self.X[feature].dtype.name != "category":
            bin_edges = np.unique(
                np.percentile(
                    original_feature_values,
                    np.linspace(0, 100, n_bins + 1),
                    interpolation="lower",
                )
            )
            # Initialize an empty ale array
            ale = np.zeros((n_bootstrap, len(bin_edges) - 1))
        else:
            # Use the unique values for discrete data.
            bin_edges = np.unique(original_feature_values)

            # Initialize an empty ale array
            ale = np.zeros((n_bootstrap, len(bin_edges)))

        # for each bootstrap set
        for k, idx in enumerate(bootstrap_indices):
            X = self.X.iloc[idx, :].reset_index(drop=True)

            # Find the ranges to calculate the local effects over
            # Using xdata ensures each bin gets the same number of X
            feature_values = X[feature].values

            # if right=True, then the smallest value in data is not included in a bin.
            # Thus, Define the bins the feature samples fall into. Shift and clip to ensure we are
            # getting the index of the left bin edge and the smallest sample retains its index
            # of 0.
            if self.X[feature].dtype.name != "category":
                indices = np.clip(
                    np.digitize(feature_values, bin_edges, right=True) - 1, 0, None
                )
            else:
                # indices for discrete data
                indices = np.digitize(feature_values, bin_edges) - 1

            # Assign the feature quantile values (based on its bin index) to two copied training datasets,
            # one for each bin edge. Then compute the difference between the corresponding predictions
            predictions = []
            for offset in range(2):
                X_temp = X.copy()
                X_temp[feature] = bin_edges[indices + offset]
                if self.estimator_output == "probability":
                    predictions.append(estimator.predict_proba(X_temp.values)[:, 1])
                elif self.estimator_output == "raw":
                    predictions.append(estimator.predict(X_temp.values))

            # The individual (local) effects.
            effects = predictions[1] - predictions[0]

            # Group the effects by their bin index
            index_groupby = pd.DataFrame(
                {"index": indices, "effects": effects}
            ).groupby("index")

            # Compute the mean local effect for each bin
            mean_effects = index_groupby.mean().to_numpy().flatten()

            # Accumulate (cumulative sum) the mean local effects.
            # Adding a 0 at the lower boundary of the first bin
            # for the interpolation step in the next step
            ale_uninterpolated = np.array([0, *np.cumsum(mean_effects)])

            # Interpolate the ale to the center of the bins.
            try:
                ale[k, :] = 0.5 * (ale_uninterpolated[1:] + ale_uninterpolated[:-1])
            except Exception as e:
                #traceback.print_exc()
                raise ValueError(
                    f"""
                                 Broadcast error!
                                 The value of n_bins ({n_bins}) is likely too 
                                 high relative to the sample size of the data. Either increase
                                 the data size (if using subsample) or use less bins. 
                                 """
                )

            # Center the ALE by substracting the bin-size weighted mean.
            ale[k, :] -= np.sum(ale[k, :] * index_groupby.size() / X.shape[0])

        results = self._store_results(
            method="ale",
            estimator_name=estimator_name,
            features=feature,
            ydata=ale,
            xdata=bin_edges,
            hist_data=[original_feature_values],
        )

        return results

    def _get_centers(self, x):
        return 0.5 * (x[1:] + x[:-1])

    def compute_second_order_ale(
        self, estimator_name, features, n_bins=30, subsample=1.0, n_bootstrap=1
    ):
        """
        Computes second-order ALE function on two continuous features data.

        Script is largely the _first_order_ale_quant from
        https://github.com/blent-ai/ALEPython/ with small modifications.

        Parameters
        ----------
        estimator_name : string
            Name of the estimator; used as a dict key

        features : string or 2-tuple of strings
            Feature name or pair of feature names to compute ICE for.

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.


        Returns
        ----------
            results : nested dictionary

        """
        estimator = self.estimators[estimator_name]

        # make sure there are two features...
        assert len(features) == 2, "Size of features must be equal to 2."

        # check to make sure both features are valid
        if features[0] not in self.feature_names:
            raise TypeError(f"Feature {features[0]} is not a valid feature")

        if features[1] not in self.feature_names:
            raise TypeError(f"Feature {features[1]} is not a valid feature")

        # create bins for computation for both features

        original_feature_values = [self.X[features[i]].values for i in range(2)]

        bin_edges = [
            np.unique(
                np.percentile(
                    v, np.linspace(0.0, 100.0, n_bins + 1), interpolation="lower"
                )
            )
            for v in original_feature_values
        ]

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X,
                subsample=subsample,
                n_bootstrap=n_bootstrap,
                seed=self.random_seed,
            )
        else:
            bootstrap_indices = [self.X.index.to_list()]

        feature1_nbin_edges = len(bin_edges[0])
        feature2_nbin_edges = len(bin_edges[1])

        ale_set = []

        # for each bootstrap set
        for k, idx in enumerate(bootstrap_indices):

            ale = np.ma.MaskedArray(
                np.zeros((feature1_nbin_edges, feature2_nbin_edges)),
                mask=np.ones((feature1_nbin_edges, feature2_nbin_edges)),
                fill_value=np.nan,
            )

            # get samples
            X = self.X.iloc[idx, :].reset_index(drop=True)

            # create bins for computation for both features
            feature_values = [X[features[i]].values for i in range(2)]

            # Define the bins the feature samples fall into. Shift and clip to ensure we are
            # getting the index of the left bin edge and the smallest sample retains its index
            # of 0.
            indices_list = [
                np.clip(np.digitize(f, b, right=True) - 1, 0, None)
                for f, b in zip(feature_values, bin_edges)
            ]

            # Invoke the predictor at the corners of the bins. Then compute the second order
            # difference between the predictions at the bin corners.
            predictions = {}
            for shifts in itertools.product(*(range(2),) * 2):
                X_temp = X.copy()
                for i in range(2):
                    X_temp[features[i]] = bin_edges[i][indices_list[i] + shifts[i]]

                if self.estimator_output == "probability":
                    predictions[shifts] = estimator.predict_proba(X_temp.values)[:, 1]
                elif self.estimator_output == "raw":
                    predictions[shifts] = estimator.predict(X_temp.values)

            # The individual (local) effects.
            effects = (predictions[(1, 1)] - predictions[(1, 0)]) - (
                predictions[(0, 1)] - predictions[(0, 0)]
            )

            # Group the effects by their indices along both axes.
            index_groupby = pd.DataFrame(
                {
                    "index_0": indices_list[0],
                    "index_1": indices_list[1],
                    "effects": effects,
                }
            ).groupby(["index_0", "index_1"])

            # Compute mean effects.
            mean_effects = index_groupby.mean()

            # Get the indices of the mean values.
            group_indices = mean_effects.index
            valid_grid_indices = tuple(zip(*group_indices))
            # Extract only the data.
            mean_effects = mean_effects.to_numpy().flatten()

            # Get the number of samples in each bin.
            n_samples = index_groupby.size().to_numpy()

            # Create a 2D array of the number of samples in each bin.
            samples_grid = np.zeros((feature1_nbin_edges - 1, feature2_nbin_edges - 1))
            samples_grid[valid_grid_indices] = n_samples

            # Mark the first row/column as valid, since these are meant to contain 0s.
            ale.mask[0, :] = False
            ale.mask[:, 0] = False

            # Place the mean effects into the final array.
            # Since `ale` contains `len(quantiles)` rows/columns the first of which are
            # guaranteed to be valid (and filled with 0s), ignore the first row and column.
            ale[1:, 1:][valid_grid_indices] = mean_effects

            # Record where elements were missing.
            missing_bin_mask = ale.mask.copy()[1:, 1:]

            # Replace the invalid bin values with the nearest valid ones.
            if np.any(missing_bin_mask):
                # Replace missing entries with their nearest neighbours.

                # Calculate the dense location matrices (for both features) of all bin centres.
                centers_list = np.meshgrid(
                    *(self._get_centers(quantiles) for quantiles in bin_edges),
                    indexing="ij",
                )

                # Select only those bin centres which are valid (had observation).
                valid_indices_list = np.where(~missing_bin_mask)
                tree = cKDTree(
                    np.hstack(
                        tuple(
                            centers[valid_indices_list][:, np.newaxis]
                            for centers in centers_list
                        )
                    )
                )

                row_indices = np.hstack(
                    [inds.reshape(-1, 1) for inds in np.where(missing_bin_mask)]
                )
                # Select both columns for each of the rows above.
                column_indices = np.hstack(
                    (
                        np.zeros((row_indices.shape[0], 1), dtype=np.int8),
                        np.ones((row_indices.shape[0], 1), dtype=np.int8),
                    )
                )

                # Determine the indices of the points which are nearest to the empty bins.
                nearest_points = tree.query(tree.data[row_indices, column_indices])[1]

                nearest_indices = tuple(
                    valid_indices[nearest_points]
                    for valid_indices in valid_indices_list
                )

                # Replace the invalid bin values with the nearest valid ones.
                ale[1:, 1:][missing_bin_mask] = ale[1:, 1:][nearest_indices]

            # Compute the cumulative sums.
            ale = np.cumsum(np.cumsum(ale, axis=0), axis=1)

            # Subtract first order effects along both axes separately.
            for i in range(2):
                # Depending on `i`, reverse the arguments to operate on the opposite axis.
                flip = slice(None, None, 1 - 2 * i)

                # Undo the cumulative sum along the axis.
                first_order = (
                    ale[(slice(1, None), ...)[flip]] - ale[(slice(-1), ...)[flip]]
                )
                # Average the diffs across the other axis.
                first_order = (
                    first_order[(..., slice(1, None))[flip]]
                    + first_order[(..., slice(-1))[flip]]
                ) / 2
                # Weight by the number of samples in each bin.
                first_order *= samples_grid
                # Take the sum along the axis.
                first_order = np.sum(first_order, axis=1 - i)
                # Normalise by the number of samples in the bins along the axis.
                first_order /= np.sum(samples_grid, axis=1 - i)
                # The final result is the cumulative sum (with an additional 0).
                first_order = np.array([0, *np.cumsum(first_order)]).reshape(
                    (-1, 1)[flip]
                )

                # Subtract the first order effect.
                ale -= first_order

            # Compute the ALE at the bin centres.
            ale = (
                reduce(
                    add,
                    (
                        ale[i : ale.shape[0] - 1 + i, j : ale.shape[1] - 1 + j]
                        for i, j in list(itertools.product(*(range(2),) * 2))
                    ),
                )
                / 4
            )

            # Center the ALE by subtracting its expectation value.
            ale -= np.sum(samples_grid * ale) / len(X)

            ale_set.append(ale)

        ###ale_set_ds = xarray.DataArray(ale_set).to_masked_array()

        results = self._store_results(
            method="ale",
            estimator_name=estimator_name,
            features=features,
            ydata=ale_set,
            xdata=bin_edges,
            hist_data=original_feature_values,
        )

        return results

    def compute_first_order_ale_cat(
        self,
        estimator_name,
        feature,
        subsample=1.0,
        n_bootstrap=1,
        feature_encoder=None,
    ):
        """
        Computes first-order ALE function on a single categorical feature.

        Script is largely from aleplot_1D_categorical from
        PyALE with small modifications (https://github.com/DanaJomar/PyALE).

        Parameters
        ----------
        estimator_name : string
            Name of the estimator; used as a dict keys

        features : string or 2-tuple of strings
            Feature name or pair of feature names to compute ICE for.

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.

        Returns
        ----------
            results : nested dictionary

        """
        pd.options.mode.chained_assignment = "raise"

        if feature_encoder is None:

            def feature_encoder_func(data):
                return data

            feature_encoder = feature_encoder_func

        estimator = self.estimators[estimator_name]
        # check to make sure feature is valid
        if feature not in self.feature_names:
            raise Exception(f"Feature {feature} is not a valid feature")

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X,
                subsample=subsample,
                n_bootstrap=n_bootstrap,
                seed=self.random_seed,
            )
        else:
            bootstrap_indices = [self.X.index.to_list()]

        original_feature_values = [feature_encoder(self.X[feature].values)]
        xdata = np.array([np.unique(original_feature_values)])
        xdata.sort()

        # Initialize an empty ale array
        ale = []

        # for each bootstrap set
        for k, idx in enumerate(bootstrap_indices):
            X = self.X.iloc[idx, :].reset_index(drop=True)

            if (X[feature].dtype.name != "category") or (not X[feature].cat.ordered):
                X[feature] = X[feature].astype(str)
                groups_order = order_groups(X, feature)
                groups = groups_order.index.values
                X[feature] = X[feature].astype(
                    pd.api.types.CategoricalDtype(categories=groups, ordered=True)
                )

            groups = X[feature].unique()
            groups = groups.sort_values()
            feature_codes = X[feature].cat.codes
            groups_counts = X.groupby(feature).size()
            groups_props = groups_counts / sum(groups_counts)

            K = len(groups)

            # create copies of the dataframe
            X_plus = X.copy()
            X_neg = X.copy()
            # all groups except last one
            last_group = groups[K - 1]
            ind_plus = X[feature] != last_group
            # all groups except first one
            first_group = groups[0]
            ind_neg = X[feature] != first_group
            # replace once with one level up
            X_plus.loc[ind_plus, feature] = groups[feature_codes[ind_plus] + 1]
            # replace once with one level down
            X_neg.loc[ind_neg, feature] = groups[feature_codes[ind_neg] - 1]
            try:
                # predict with original and with the replaced values
                # encode the categorical feature
                X_coded = pd.concat(
                    [
                        X.drop(feature, axis=1),
                        feature_encoder(X[[feature]]),
                    ],
                    axis=1,
                )
                X_coded = np.array(X_coded[self.feature_names], dtype=float)

                # predict
                if self.estimator_output == "probability":
                    y_hat = estimator.predict_proba(X_coded)[:, 1]
                else:
                    y_hat = estimator.predict(X_coded)

                # encode the categorical feature
                X_plus_coded = pd.concat(
                    [
                        X_plus.drop(feature, axis=1),
                        feature_encoder(X_plus[[feature]]),
                    ],
                    axis=1,
                )

                X_plus_coded = np.array(
                    X_plus_coded[ind_plus][self.feature_names], dtype=float
                )

                # predict
                if self.estimator_output == "probability":
                    y_hat_plus = estimator.predict_proba(X_plus_coded)[:, 1]
                else:
                    y_hat_plus = estimator.predict(X_plus_coded)

                # encode the categorical feature
                X_neg_coded = pd.concat(
                    [
                        X_neg.drop(feature, axis=1),
                        feature_encoder(X_neg[[feature]]),
                    ],
                    axis=1,
                )

                X_neg_coded = np.array(
                    X_neg_coded[ind_neg][self.feature_names], dtype=float
                )

                # predict
                if self.estimator_output == "probability":
                    y_hat_neg = estimator.predict_proba(X_neg_coded)[:, 1]
                else:
                    y_hat_neg = estimator.predict(X_neg_coded)

            except Exception as ex:
                raise Exception(
                    """There seems to be a problem when predicting with the estimator.
                    Please check the following: 
                    - Your estimator is fitted.
                        - The list of predictors contains the names of all the features"""
                    """ used for training the estimator.
                        - The encoding function takes the raw feature and returns the"""
                    """ right columns encoding it, including the case of a missing category.
                    """
                )

            # compute prediction difference
            Delta_plus = y_hat_plus - y_hat[ind_plus]
            Delta_neg = y_hat[ind_neg] - y_hat_neg

            # compute the mean of the difference per group
            delta_df = pd.concat(
                [
                    pd.DataFrame(
                        {
                            "eff": Delta_plus,
                            feature: groups[feature_codes[ind_plus] + 1],
                        }
                    ),
                    pd.DataFrame(
                        {"eff": Delta_neg, feature: groups[feature_codes[ind_neg]]}
                    ),
                ]
            )

            res_df = delta_df.groupby([feature]).mean()
            res_df.loc[:, "ale"] = res_df.loc[:, "eff"].cumsum()

            res_df.loc[groups[0]] = 0
            # sort the index (which is at this point an ordered categorical) as a safety measure
            res_df = res_df.sort_index()

            # Subtract the mean value to get the centered value.
            ale_temp = res_df["ale"] - sum(res_df["ale"] * groups_props)
            ale.append(ale_temp)

        ale = np.array(ale, dtype=float)

        results = self._store_results(
            method="ale",
            estimator_name=estimator_name,
            features=feature,
            ydata=ale,
            xdata=xdata,
            hist_data=original_feature_values,
            categorical=True,
        )

        return results

    def compute_scalar_interaction_stats(
        self,
        method,
        data,
        estimator_names,
        data_2d=None,
        features=None,
        **kwargs,
    ):
        """
        Wrapper function for computing the interaction strength statistic or the Friedman
        H-statistic (see below). Will perform calculation in parallel for multiple estimators.

        Parameters
        ------------

        method : 'ias' or 'hstat'
            Whether to compute the interaction strength (ias) or the
            Friedman H-statistics (hstat)

        data : xarray.Dataset
        data_2d : xarray.Dataset

        features : string

        estimator_name : string
            Name of the estimator; used as a dict keys

        n_bins : integer
            Number of bins

        subsample: float or integer (default=1.0 for no subsampling)

            if value is between 0-1, it is interpreted as fraction of total X to use
            if value > 1, interpreted as the number of X to randomly sample
            from the original dataset.

        n_bootstrap: integer (default=None for no bootstrapping)
            number of bootstrap resamples for computing confidence intervals.

        n_jobs : integer or float

        Returns
        --------
            results : xarray.Dataset

        """
        if method == "ias":
            func = self.compute_interaction_strength
            args_iterator = to_iterator(
                estimator_names,
            )
            total = len(estimator_names)
        elif method == "hstat":
            func = self.friedman_h_statistic
            args_iterator = to_iterator(
                estimator_names,
                features,
            )
            total = len(estimator_names) * len(features)

        self.data = data
        self.data_2d = data_2d
        n_jobs = len(estimator_names)
        results = run_parallel(
            func=func,
            args_iterator=args_iterator,
            kwargs=kwargs,
            nprocs_to_use=n_jobs,
            total=total,
        )

        results = merge_dict(results)

        if method == "hstat":
            final_results = {}
            feature_pairs = np.array([f"{f[0]}__{f[1]}" for f in features])
            for estimator_name in estimator_names:
                values = np.array(
                    [results[f"{f}__{estimator_name}_hstat"] for f in feature_pairs]
                )
                idx = np.argsort(np.mean(values, axis=1))[::-1]
                feature_names_sorted = np.array(feature_pairs)[idx]
                values_sorted = values[idx, :]

                final_results[f"hstat_rankings__{estimator_name}"] = (
                    [f"n_vars_perm_based_interactions"],
                    feature_names_sorted,
                )
                final_results[f"hstat_scores__{estimator_name}"] = (
                    [f"n_vars_perm_based_interactions", "n_bootstrap"],
                    values_sorted,
                )
            results = to_xarray(final_results)
        else:
            results = to_xarray(results)

        return results

    def friedman_h_statistic(self, estimator_name, features, **kwargs):
        """
        Compute the H-statistic for two-way interactions between two features.

        Parameters
        -------------
        estimator_name : str
        feature_tuple : 2-tuple of strs
        n_bins : int
        subsample : integer or float
        n_bootstrap : integer

        Returns
        --------

        results : dictionary
        """
        feature1, feature2 = features
        feature1_pd = self.data[f"{feature1}__{estimator_name}__pd"].values
        feature2_pd = self.data[f"{feature2}__{estimator_name}__pd"].values

        combined_pd = self.data_2d[
            f"{feature1}__{feature2}__{estimator_name}__pd"
        ].values

        # Calculate the H-statistics
        pd_decomposed = feature1_pd[:, :, np.newaxis] + feature2_pd[:, np.newaxis, :]

        numer = (combined_pd - pd_decomposed) ** 2
        denom = (combined_pd) ** 2
        H_squared = np.sum(np.sum(numer, axis=2), axis=1) / np.sum(
            np.sum(denom, axis=2), axis=1
        )

        results = {
            f"{feature1}__{feature2}__{estimator_name}_hstat": np.sqrt(H_squared)
        }

        return results

    def compute_interaction_strength(self, estimator_name, **kwargs):
        """
        Compute the interaction strenth of a ML estimator (based on IAS from
        Quantifying estimator Complexity via Functional
        Decomposition for Better Post-Hoc Interpretability).

        Parameters
        --------------------
        estimator_names : list of strings

        Returns
        ---------

        ias : dict

        """
        subsample = kwargs.get("subsample", 1.0)
        n_bootstrap = kwargs.get("n_bootstrap", 1)
        estimator_output = kwargs.get("estimator_output", "raw")

        estimator = self.estimators[estimator_name]
        feature_names = list(self.X.columns)
        data = self.data

        # Get the interpolated ALE curves
        main_effect_funcs = {}
        for f in feature_names:
            ale = np.mean(data[f"{f}__{estimator_name}__ale"].values, axis=0)
            x = data[f"{f}__bin_values"].values
            main_effect_funcs[f] = interp1d(
                x,
                ale,
                fill_value="extrapolate",
                kind="linear",
            )

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X, subsample=subsample, n_bootstrap=n_bootstrap
            )
        else:
            bootstrap_indices = [self.X.index.to_list()]

        ias = []
        for k, idx in enumerate(bootstrap_indices):
            X = self.X.iloc[idx, :].values
            if self.estimator_output == "probability":
                predictions = estimator.predict_proba(X)[:, 1]
            else:
                predictions = estimator.predict(X)

            # Get the average estimator prediction
            avg_prediction = np.mean(predictions)

            # Get the ALE value for each feature per X
            main_effects = np.array(
                [
                    main_effect_funcs[f](np.array(X[:, i], dtype=np.float64))
                    for i, f in enumerate(feature_names)
                ]
            )

            # Sum the ALE values per X and add on the average value
            main_effects = np.sum(main_effects.T, axis=1) + avg_prediction

            num = np.sum((predictions - main_effects) ** 2)
            denom = np.sum((predictions - avg_prediction) ** 2)

            # Compute the interaction strength
            ias.append(num / denom)

        return {f"{estimator_name}_ias": (["n_bootstrap"], np.array(ias))}

    def compute_ale_variance(self, data, estimator_names, features=None, **kwargs):
        """
        Compute the standard deviation of the ALE values
        for each feature and rank then for predictor importance.

        Parameters
        ----------
        data : xarray.Dataset
        estimator_names : list of strings
        features : str
        ale_subsample : float or integer

        Returns
        ---------

        results_ds : xarray.Dataset

        """
        feature_names = list([f for f in data.data_vars if "__" not in f])
        feature_names.sort()

        features, cat_features = determine_feature_dtype(self.X, feature_names)

        def _std(values, f, cat_features):
            """
            Using a different formula for computing standard deviation for
            categorical features.
            """
            if f in cat_features:
                return 0.25 * (np.max(values, axis=1) - np.min(values, axis=1))
            else:
                return np.std(values, ddof=1, axis=1)

        results = {}
        for estimator_name in estimator_names:
            # Compute the std over the bin axis [shape = (n_features, n_bootstrap)]
            # Input shape : (n_bootstrap, n_bins)
            ale_std = np.array(
                [
                    _std(data[f"{f}__{estimator_name}__ale"].values, f, cat_features)
                    for f in feature_names
                ]
            )

            # Average over the bootstrap indices
            idx = np.argsort(np.mean(ale_std, axis=1))[::-1]

            feature_names_sorted = np.array(feature_names)[idx]
            ale_std_sorted = ale_std[idx, :]

            results[f"ale_variance_rankings__{estimator_name}"] = (
                [f"n_vars_ale_variance"],
                feature_names_sorted,
            )
            results[f"ale_variance_scores__{estimator_name}"] = (
                [f"n_vars_ale_variance", "n_bootstrap"],
                ale_std_sorted,
            )

        results_ds = to_xarray(results)

        return results_ds

    def compute_interaction_rankings(
        self,
        data,
        estimator_names,
        features,
        **kwargs,
    ):
        """
        Compute the variable interaction rankings from Greenwell et al. 2018, but
        using the purely second-order ALE rather than the second-order PD.

        For a given second-order ALE,
         1. Compute the standard deviation over one axis and then take the std of those values
         2. Compute the standard deviation over another axis and then take the std of those values
         3. Average 1 and 2 together.

        Sort by magnitudes given in (3). Larger values (relative to other values) are
        possibly indicative of  stronger feature interactions.


        Parameters
        --------------------
        data : xarray.Dataset
        estimator_names : list of strings
        features : str
        ale_subsample : float or integer

        Returns
        --------
        results_ds : xarray.Dataset
        """
        results = {}
        for estimator_name in estimator_names:

            interaction_effects = []
            for f in features:
                data_temp = data[f"{f[0]}__{f[1]}__{estimator_name}__ale"].values
                std_feature0 = np.std(np.std(data_temp, ddof=1, axis=1), ddof=1, axis=1)
                std_feature1 = np.std(np.std(data_temp, ddof=1, axis=2), ddof=1, axis=1)
                interaction_effects.append(0.5 * (std_feature0 + std_feature1))

            interaction_effects = np.array(interaction_effects)
            # Average over the bootstrap indices
            idx = np.argsort(np.mean(interaction_effects, axis=1))[::-1]

            feature_names_sorted = np.array(features)[idx]
            interaction_effects_sorted = interaction_effects[idx, :]
            feature_names_sorted = [f"{f[0]}__{f[1]}" for f in feature_names_sorted]

            results[f"ale_variance_interactions_rankings__{estimator_name}"] = (
                [f"n_vars_ale_variance_interactions"],
                feature_names_sorted,
            )
            results[f"ale_variance_interactions_scores__{estimator_name}"] = (
                [f"n_vars_ale_variance_interactions", "n_bootstrap"],
                interaction_effects_sorted,
            )

        results_ds = to_xarray(results)

        return results_ds

    def number_of_features_used(self, estimator, X):
        """
        Compute the number of features (NF) from Molnar et al. 2019
        """
        original_predictions = estimator.predict_proba(X)[:, 1]
        features = list(X.columns)
        for f in features:
            X_temp = X.copy()
            X_temp.loc[:, f] = np.random.permutation(X_temp[f].values)

            new_predictions = estimator.predict_proba(X_temp)[:, 1]
            change = np.absolute(new_predictions - original_predictions)

    def compute_interaction_performance_based(
        self,
        estimator,
        X,
        y,
        X_permuted,
        features,
        evaluation_fn,
        estimator_output,
        verbose=False,
    ):
        """
        Compute the performance-based feature interactions from Oh (2019).

        Err(F_i) : Error when feature F_i is permuted as compared
               against the original estimator performance.

        Err({F_i, F_j}) : Error when features F_i & F_j are permuted as
                       compared against the original estimator performance.

        Interaction = Err(F_i) + Err(F_j) - Err({F_i, F_j}), for classification
        Interaction = Err({F_i, F_j}) - (Err(F_i) + Err(F_j)), for regression

        Interaction(F_i, F_j) = 0 -> no interaction between F_i & F_j
        Interaction(F_i, F_j) > 0 -> positive interaction between F_i & F_j
        Interaction(F_i, F_j) < 0 -> negative interaction between F_i & F_j

        Negative interaction implies that the connection between Fi and Fj reduces the prediction performance,
        whereas positive interaction leads to an increase in the prediction performance.
        In other words, positive or negative interactions decrease or increase the
        prediction error, respectively.

        Parameters
        -----------
        estimator,
        X : pandas.DataFrame  of shape (n_X, n_features)
        Data to compute importance over.
        y : numpy.array
        y values
        features: list of 2-tuples
        evaluation_fn : callable
        random_state : integer

        Returns
        --------

        err : numpy.array
        """
        available_scores = ["auc", "auprc", "bss", "mse", "norm_aupdc"]

        if isinstance(evaluation_fn, str):
            evaluation_fn = evaluation_fn.lower()
            is_str = True

        if isinstance(evaluation_fn, str):
            if evaluation_fn == "auc":
                evaluation_fn = roc_auc_score
                scoring_strategy = "argmin_of_mean"
            elif evaluation_fn == "auprc":
                evaluation_fn = average_precision_score
                scoring_strategy = "argmin_of_mean"
            elif evaluation_fn == "norm_aupdc":
                evaluation_fn = norm_aupdc
                scoring_strategy = "argmin_of_mean"
            elif evaluation_fn == "bss":
                evaluation_fn = brier_skill_score
                scoring_strategy = "argmin_of_mean"
            elif evaluation_fn == "mse":
                evaluation_fn = mean_squared_error
                scoring_strategy = "argmax_of_mean"
            else:
                raise ValueError(
                    f"evaluation_fn is not set! Available options are {available_scores}"
                )

        if estimator_output == "probability":
            # Classification
            predictions = estimator.predict_proba(X)[:, 1]
        else:
            # Regression
            predictions = estimator.predict(X)

        original_score = evaluation_fn(y, predictions)

        err = 0
        X_permuted_both = X.copy()
        # Compute the change in estimator performance for the two features separately
        for feature in features:
            X_temp = X.copy()
            X_temp.loc[:, feature] = X_permuted.loc[:, feature]
            X_permuted_both.loc[:, feature] = X_permuted.loc[:, feature]

            # Get the predictions for the dataset with a permuted feature.
            if estimator_output == "probability":
                # Classification
                predictions = estimator.predict_proba(X_temp)[:, 1]
            else:
                # Regression
                predictions = estimator.predict(X_temp)

            # Compute the permuted score
            permuted_score = evaluation_fn(y, predictions)

            if estimator_output == "probability":
                # Classification
                err_i = original_score - permuted_score
            else:
                # Regression
                err_i = permuted_score - original_score

            if verbose:
                permuted_features = check_is_permuted(X, X_temp)
                print(f"Reduced performance by {feature} : {err_i}")
                print(f"Permuted Features: {permuted_features}")

            err += err_i

        # Compute the change in estimator performance with both features permuted.
        if estimator_output == "probability":
            predictions_both = estimator.predict_proba(X_permuted_both)[:, 1]
        else:
            predictions_both = estimator.predict(X_permuted_both)

        # Compute the permuted score
        permuted_score_both = evaluation_fn(y, predictions_both)

        if estimator_output == "probability":
            err_both = original_score - permuted_score_both
        else:
            err_both = permuted_score_both - original_score

        if verbose:
            permuted_features = check_is_permuted(X, X_permuted_both)
            print(f"Reduced performance by {features} : {err_both}")
            print(f"Permuted Features: {permuted_features}")

        # Combine for the feature interaction between the two features
        if estimator_output == "probability":
            err -= err_both
        else:
            err = err_both - err

        return err

    def compute_interaction_rankings_performance_based(
        self,
        estimator_names,
        features,
        evaluation_fn,
        estimator_output,
        subsample=1.0,
        n_bootstrap=1,
        n_jobs=1,
        verbose=False,
    ):
        """
        Wrapper function for performance_based_feature_interactions
        """
        unique_features = list(set([item for t in features for item in t]))
        n_feature_pairs = len(features)
        X_permuted = self.X.copy()
        # Permute all features up front to save on computation cost
        for f in unique_features:
            X_permuted.loc[:, f] = np.random.permutation(self.X.loc[:, f])

        args_iterator = to_iterator(
            estimator_names,
            features,
            [X_permuted],
            [evaluation_fn],
            [estimator_output],
            [subsample],
            [n_bootstrap],
            [verbose],
        )

        results = run_parallel(
            func=self._feature_interaction_worker,
            args_iterator=args_iterator,
            kwargs={},
            nprocs_to_use=n_jobs,
        )

        results = merge_dict(results)

        final_results = {}
        feature_pairs = np.array([f"{f[0]}__{f[1]}" for f in features])
        for estimator_name in estimator_names:
            values = np.array(
                [results[f"{f}_interaction__{estimator_name}"] for f in feature_pairs]
            )
            idx = np.argsort(np.absolute(np.mean(values, axis=1)))[::-1]
            feature_names_sorted = np.array(feature_pairs)[idx]
            values_sorted = values[idx, :]

            final_results[f"perm_based_interactions_rankings__{estimator_name}"] = (
                [f"n_vars_perm_based_interactions"],
                feature_names_sorted,
            )
            final_results[f"perm_based_interactions_scores__{estimator_name}"] = (
                [f"n_vars_perm_based_interactions", "n_bootstrap"],
                values_sorted,
            )

        results_ds = to_xarray(final_results)

        return results_ds

    def _feature_interaction_worker(
        self,
        estimator_name,
        features,
        X_permuted,
        evaluation_fn,
        estimator_output,
        subsample=1.0,
        n_bootstrap=1,
        verbose=False,
    ):
        """
        Internal worker function for parallel computations.
        """
        estimator = self.estimators[estimator_name]

        # get the bootstrap samples
        if n_bootstrap > 1 or float(subsample) != 1.0:
            bootstrap_indices = compute_bootstrap_indices(
                self.X, subsample=subsample, n_bootstrap=n_bootstrap
            )
        else:
            bootstrap_indices = [np.arange(self.X.shape[0])]

        results = {}
        err_set = np.zeros((n_bootstrap,))
        for k, idx in enumerate(bootstrap_indices):
            # get samples
            X = self.X.iloc[idx, :].reset_index(drop=True)
            y = self.y[idx]

            err = self.compute_interaction_performance_based(
                estimator,
                X,
                y,
                X_permuted,
                features,
                evaluation_fn,
                estimator_output,
                verbose=verbose,
            )
            err_set[k] = err

        results[f"{features[0]}__{features[1]}_interaction__{estimator_name}"] = err_set

        return results

    def compute_main_effect_complexity(
        self,
        estimator_name,
        ale_ds,
        features,
        post_process=False,
        max_segments=10,
        approx_error=0.05,
        debug=False,
    ):
        """
        Compute the Main Effect Complexity (MEC; Molnar et al. 2019).
        MEC is the number of linear segements required to approximate
        the first-order ALE curves; averaged over all features.
        The MEC is weighted-averged by the variance. Higher values indicate
        a more complex estimator (less interpretable).

        References
        -----------
            Molnar, C., G. Casalicchio, and B. Bischl, 2019: Quantifying estimator Complexity via
            Functional Decomposition for Better Post-Hoc Interpretability. ArXiv.


        Parameters
        ----------------

        ale_ds : xarray.Dataset

            The results xarray.Dataset from computing 1D ALE using .calc_ale().
            Must be computed for all features in X.

        estimator_names : string, list of strings

            If using multiple estimators, you can pass a single (or subset of) estimator name(s)
            to compute the MEC for.

        max_segments : integer

            Maximum number of linear segments used to approximate the main/first-order
            effect of a feature. default is 10. Used to limit the computational expense.

        approx_error : float

            The accepted error of the R squared between the piece-wise linear function
            and the true ALE curve. If the R square is within the approx_error, then
            no additional segments are added. default is 0.05

        Returns
        ----------
            mec_avg, float
                Average Main Effect Complexity
            best_break_dict, dict
                For each feature, the list of "optimal" breakpoints
                Used to plot and verify the code is running correctly.
        """
        mec = np.zeros((len(features)))
        var = np.zeros((len(features)))

        # Based on line 6 from Algorithm 2: Main Effect Complexity (MEC) in Molnar et al. 2019
        # we ignore categorical features (set the slopes to zero).
        categorical_features = [
            self.X[f].dtype.name == "category" for f in self.feature_names
        ]
        best_breaks_dict = {}
        best_g = {}
        for j, f in enumerate(features):
            ale = np.mean(ale_ds[f"{f}__{estimator_name}__ale"].values, axis=0)
            x = ale_ds[f"{f}__bin_values"].values.reshape(-1, 1)

            b_max = len(x)
            # Approximate ALE with linear estimator
            lr = LinearRegression()
            lr.fit(x.reshape(-1, 1), ale)
            g = lr.predict(x)

            best_g[f] = g

            coef = lr.coef_[0]
            best_score = r2_score(g, ale)

            # Increase num. of segements until approximation is good enough.
            k = 1
            best_breaks = []
            while k < max_segments and (r2_score(g, ale) < (1 - approx_error)):
                # Find intervals Z_k through exhaustive search along ALE curve breakpoints
                for b in range(1, b_max - 1):
                    if b not in best_breaks:
                        temp_breaks = copy(best_breaks)
                        temp_breaks.append(b)
                        temp_breaks.sort()

                        idxs_set = (
                            [range(0, temp_breaks[0] + 1)]
                            + [
                                range(temp_breaks[i], temp_breaks[i + 1] + 1)
                                for i in range(len(temp_breaks) - 1)
                            ]
                            + [range(temp_breaks[-1], b_max)]
                        )

                        estimator_set = [
                            LinearRegression(normalize=True)
                            for _ in range(len(idxs_set))
                        ]

                        estimator_fit_set_tmp = [
                            estimator_set[i].fit(x[idxs, :], ale[idxs])
                            for i, idxs in enumerate(idxs_set)
                        ]

                        # For categorical features, set the slope to zero (but keep the intercept).
                        if f in categorical_features:
                            estimator_fit_set = []
                            for e in estimator_fit_set_tmp:
                                e.coef_ = np.array([0.0])
                                estimator_fit_set.append(e)
                        else:
                            estimator_fit_set = [e for e in estimator_fit_set_tmp]

                        predict_set = [
                            estimator_fit_set[i].predict(x[idxs, :])
                            for i, idxs in enumerate(idxs_set)
                        ]

                        # Combine the predictions from the different breaks into
                        # a single piece-wise function (line 7 for Molnar et al. 2019)
                        for i, idxs in enumerate(idxs_set):
                            g[idxs] = predict_set[i]

                        current_score = r2_score(g, ale)
                        if current_score > best_score:
                            best_score = current_score
                            best_break = b
                            # Is approx good enough? Then stop iterating
                            if best_score > 1 - approx_error:
                                best_g[f] = g
                                break

                best_breaks.append(best_break)
                k += 1

            # Sum of non-zero coefficients minus first intercept
            best_breaks_dict[f] = best_breaks
            mec[j] = k  # + add
            var[j] = np.var(ale)

        mec_avg = np.average(mec, weights=var)

        if debug:
            return mec_avg, best_breaks_dict, best_g
        else:
            return mec_avg, best_breaks_dict

    def scorer(self, estimator, X, y, evaluation_fn):
        if hasattr(estimator, 'predict_proba'):
            prediction = estimator.predict_proba(X)[:, 1]
        elif hasattr(estimator, 'predict'):
            prediction = estimator.predict(X)[:]
        else:
            raise AttributeError(f'{estimator} does not have .predict or .predict_proba!')
            
        return evaluation_fn(y, prediction)

    def _compute_grouped_importance(
        self,
        X,
        y,
        evaluation_fn,
        group,
        feature_subset,
        estimator,
        only,
        all_permuted_score,
        scoring_strategy, 
    ):
        scores = {group: []}
        X_permuted = X.copy()
        for rs in self.random_states:
            inds = rs.permutation(len(X))
            # Jointly permute all features expect those in the feature_subset

            if only:
                # For group only version, jointly permute all features except those in the feature subset.
                X_permuted = np.array(
                    [
                        X[:, i] if i in feature_subset else X[inds, i]
                        for i in range(X.shape[1])
                    ]
                ).T
            else:
                # Else, only jointly permute the features in the feature subset.
                X_permuted = np.array(
                    [
                        X[inds, i] if i in feature_subset else X[:, i]
                        for i in range(X.shape[1])
                    ]
                ).T

            group_permute_score = self.scorer(estimator, X_permuted, y, evaluation_fn)

            if only:
                # For metrics like AUC, NAUPDC, CSI, BSS
                if scoring_strategy == 'minimize':
                    imp = group_permute_score - all_permuted_score
                else:
                    imp = all_permuted_score - group_permute_score
                    
            else:
                if scoring_strategy == 'minimize':
                    imp = all_permuted_score - group_permute_score
                else:
                    imp = group_permute_score - all_permuted_score

            scores[group].append(imp)

        return scores

    def grouped_feature_importance(
        self,
        evaluation_fn,
        perm_method,
        scoring_strategy=None, 
        n_permute=1,
        groups=None,
        sample_size=100,
        subsample=1.0,
        n_jobs=1,
        clustering_kwargs={"n_clusters": 10},
    ):
        """
        The group only permutation feature importance (GOPFI) from Au et al. 2021
        (see their equations 10,11). This function has a built-in method for clustering
        features using the sklearn.cluster.FeatureAgglomeration. It also has the ability to
        compute the results over multiple permutations to improve the feature importance
        estimate (and provide uncertainty).

        Description of the parameters provided in ExplainToolkit.

        """
        parallel = Parallel(n_jobs=n_jobs)
        only = True if perm_method == "grouped_only" else False

        if isinstance(evaluation_fn, str):
            evaluation_fn, scoring_strategy = self._to_scorer(evaluation_fn)
        else:
            if scoring_strategy is None:
                raise ValueError(
                """ 
                The scoring_strategy argument is None! If you are using a non-default evaluation_fn 
                then scoring_strategy must be set! If the metric is positively-oriented (a higher value is better), 
                then set scoring_strategy = "minimize" (i.e., a lower score means greater importance) 
                and if it is negatively-oriented (a lower value is better), then set scoring_strategy = "maximize"
                """
            )
        if scoring_strategy == 'argmin_of_mean':
            scoring_strategy = 'minimize'
        elif scoring_strategy == 'argmax_of_mean':
            scoring_strategy = 'maximize'
                
        self.random_states = bootstrap_generator(n_bootstrap=n_permute)

        # Get the feature names from the dataframe.
        feature_names = np.array(self.X.columns)

        # Convert to numpy array
        X = self.X.values

        subsample = subsample if subsample > 1 else int(subsample * X.shape[0])
        inds = self.random_states[-1].choice(len(X), size=subsample, replace=False)

        # Subsample the examples for computational efficiency.
        X = X[inds]
        y = self.y[inds]

        if groups is None:
            # Feature clustering is based on a subset of examples for computational efficiency.
            sample_size = (
                sample_size if sample_size > 1 else int(sample_size * X.shape[0])
            )
            inds = np.random.choice(len(X), size=sample_size, replace=False)
            agglo = cluster.FeatureAgglomeration(**clustering_kwargs)
            agglo.fit(X[inds, :])

            groups = {
                f"group {i}": np.where(agglo.labels_ == i)[0]
                for i in range(np.max(agglo.labels_) + 1)
            }
            names = {
                f"group {i}": feature_names[agglo.labels_ == i]
                for i in range(np.max(agglo.labels_) + 1)
            }
        else:
            contains_str = isinstance(list(groups.values())[0][0], str)
            if contains_str:
                names = copy(groups)
                # It is the feature names and thus need to be converted to indices
                for key, items in groups.items():
                    N = np.where(np.isin(feature_names, items))[0]
                    groups[key] = N
            else:
                names = {key: feature_names[inds] for key, inds in groups.items()}

        X_permuted = X.copy()
        if only:
            # for the group only version, we jointly permute all features and then
            # determine the original score
            inds = np.random.permutation(len(X))
            X_permuted = np.array([X[inds, i] for i in range(X.shape[1])]).T

        results = []
        for estimator_name, estimator in self.estimators.items():
            # Score after jointly permuting all features.
            all_permuted_score = self.scorer(estimator, X_permuted, y, evaluation_fn)

            scores = parallel(
                delayed(self._compute_grouped_importance)(
                    X,
                    y,
                    evaluation_fn,
                    group,
                    feature_subset,
                    estimator,
                    only,
                    all_permuted_score,
                    scoring_strategy,
                )
                for group, feature_subset in groups.items()
            )
            scores = merge_dict(scores)

            group_names = list(groups.keys())
            importances = np.array([scores[g] for g in group_names])

            group_rank = to_skexplain_importance(
                importances,
                estimator_name=estimator_name,
                feature_names=group_names,
                method=perm_method,
            )

            results.append(group_rank)

        results = xr.merge(results, combine_attrs="override", compat="override")

        return results, names
